#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <exec/exec_protos.h>
#include <dos/dos.h>

#define MAX_ARG_LEN 256

#define F_ALIGN_LEFT         1
#define F_ALWAYS_SIGN        2
#define F_NEGATIVE           4
#define F_SPACE_IF_NOT_SIGN  8
#define F_FILL_ZERO         16
#define F_ALTERNATE         32


#define GETNEXT(a, b) {if(!((a) = (*((b)++)))) {va_end(ap); return -1;}}

static const char hex[16]="0123456789abcdef";

static int add(char *src, int n, int w, ULONG flags, int (*cf)(int, void *),
	       void *fp)
{
    int i;
    char leftfillchar = (flags &  F_FILL_ZERO) ? '0' : ' ';
    int acwritten;

    w -= n;
    if (flags & (F_NEGATIVE | F_ALWAYS_SIGN | F_SPACE_IF_NOT_SIGN))
	w--;

    acwritten = 0;
    if (!(flags & F_ALIGN_LEFT) && leftfillchar == ' ')
    {
	if (w > 0)
	{
	    acwritten += w;
	    while (w-- > 0)
		cf(' ', fp);
	}
    }

    if (flags & (F_NEGATIVE | F_ALWAYS_SIGN | F_SPACE_IF_NOT_SIGN))
    {
	acwritten++;
	if (flags & F_NEGATIVE)
	{
	    cf('-', fp);
	}
	else
	{
	    if (flags & F_ALWAYS_SIGN)
		cf('+', fp);
	    else
		cf(' ', fp);
	}
    }

    if (!(flags & F_ALIGN_LEFT) && leftfillchar == '0')
	if (w > 0)
	{
	    acwritten += w;
	    while (w-- > 0)
		cf(leftfillchar, fp);
	}

    acwritten += n;
    for (i=0; i<n; i++)
        cf(*src++, fp);

    if (flags & F_ALIGN_LEFT)
	if (w > 0)
	{
	    acwritten += w;
	    while (w-- > 0)
		cf(' ', fp);
	}

    return acwritten;
}



LONG _PrintFmt(va_list ap, void *fp, int (*cf)(int, void *), const char *fmt)
{
    long value;
    ULONG uvalue;
    char *svalue;
    char curr;
    int written = 0;

    curr = *fmt++;
    while (curr)
    {
	if (curr == '%')
	{
	    BOOL currucase;
	    int width = 0;
	    char tmpstr[20];
	    int i;
	    ULONG flags = 0;
	    BOOL phase1 = TRUE;
	    BOOL dont_get = FALSE;

	    while (phase1) {
		if (!dont_get)
		    GETNEXT(curr, fmt);
		dont_get = FALSE;

		switch(curr)
		{
		case '-':
		    flags |= F_ALIGN_LEFT;
		    break;
		case '+':
		    flags |= F_ALWAYS_SIGN;
		    break;
		case ' ':
		    flags |= F_SPACE_IF_NOT_SIGN;
		    break;
		case '0':
		    flags |= F_FILL_ZERO;
		    break;
		case '#':
		    flags |= F_ALTERNATE;
		    break;
		default:
		    if (isdigit(curr))
		    {
			while (isdigit(curr))
			{
			    width = width*10 + (curr - '0');
			    GETNEXT(curr, fmt);
			}
			dont_get = TRUE;
		    }
		    else
		    {
			phase1 = FALSE;
		    }
		    break;
		}		
	    }

	    if (tolower(curr) == 'l' || curr == 'h')
		GETNEXT(curr, fmt);

	    currucase = isupper(curr);
	    switch(tolower(curr)) {
	    case 'i':
	    case 'd':
		value = va_arg(ap, long);
		i = 0;
		if (value == 0x80000000)
		{
		    flags |= F_NEGATIVE;
		    memcpy(&tmpstr[19-(10-1)-i], "2147483648", 10);
		    i += 10;
		}
		else
		{
		    if (value < 0)
		    {
			flags |= F_NEGATIVE;
			value = -value;
		    }
		    while (value)
		    {
			tmpstr[19-i++] = value % 10 + '0';
			value /= 10;
		    }
		}
		if (!i) tmpstr[19-i++] = '0';
		written += add(tmpstr+20-i, i, width, flags, cf, fp);
		break;
	    case 'u':
		uvalue = va_arg(ap, long);
		i = 0;
		while (uvalue)
		{
		    tmpstr[19-i++] = uvalue % 10 + '0';
		    uvalue /= 10;
		}
		if (!i) tmpstr[19-i++] = '0';
		written += add(tmpstr+20-i, i, width, flags, cf, fp);
		break;
	    case 'o':
		uvalue = va_arg(ap, long);
		i = 0;

		while (uvalue)
		{
		    tmpstr[19-i++] = (uvalue & 7) + '0';
		    uvalue >>= 3;
		}
		if (!i || flags & F_ALTERNATE)
		    tmpstr[19-i++] = '0';

		written += add(tmpstr+20-i, i, width, flags, cf, fp);
		break;
	    case 'p':
	    case 'x':
		uvalue = va_arg(ap, unsigned long);
		i = 0;
		while (uvalue)
		{
		    tmpstr[19-i++] = (currucase) ?
			toupper(hex[uvalue & 15]) :
			hex[uvalue & 15];
		    uvalue >>= 4;
		}
		if (i && flags & F_ALTERNATE)
		{
		    tmpstr[19-i++] = curr;
		    tmpstr[19-i++] = '0';
		}
		if (!i) tmpstr[19-i++] = '0';

		written += add(tmpstr+20-i, i, width, flags, cf, fp);
		break;
	    case 'c':
		tmpstr[0] = va_arg(ap, char);
		add(tmpstr, 1, 1, 0, cf, fp);
		break;
	    case 's':
		svalue = va_arg(ap, char *);
		flags &= ~(F_FILL_ZERO | F_ALWAYS_SIGN | F_SPACE_IF_NOT_SIGN);
		written += add(svalue, strlen(svalue), width, flags, cf, fp);
		break;
	    case 'n':
		*(va_arg(ap, int *)) = written;
		break;
	    case 'f':
	    case 'e':
	    case 'g':
		written += add("<float>", 7, 7, 0, cf, fp);
		break;
	    case '%':
		tmpstr[0] = '%';
		add(tmpstr, 1, 1, 0, cf, fp);
		break;
	    default:
		return -1;   /* Unrecognized code, aborting */
		break;
	    }
	}
	else
	{
	    written++;
	    cf(curr, fp);
	}

	curr = *fmt++;
    }

    va_end(ap);

    return written;
}



static int FPutc(int c, void *fp)
{
    char a = c;

    return Write((APTR)fp, &a, 1);
}

LONG _FPrintF(APTR fp, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);

    return PrintFmt(ap, (void *)fp, FPutc, fmt);
}

LONG _PrintF(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);

    return PrintFmt(ap, (void *)Output(), FPutc, fmt);
}



struct ArgList *_ParseArgs(const char *args)
{
    struct ArgList *AL;

    AL = AllocMem(sizeof(struct ArgList), 0);
    if (!AL)
	return NULL;

    NewList(&AL->ArgMemList);
    AL->argc = 0;
    AL->argv = NULL;

    if (args)
    {
	int argc = 0;
	char **argv;
	char argtmp[MAX_ARG_LEN], *currarg = NULL;
	int maxargs = 4, currarglen = 0;
	BOOL in_word = FALSE, in_par = FALSE;

	argv = AllocRemember(&AL->ArgMemList, maxargs * sizeof(char *), 0);
	if (!argv)
	{
	    FreeArgs(AL);
	    return NULL;
	}

	do
	{
	    if ((isspace(*args) && !in_par) || !(*args)) /* if space or NULL */
	    {
		if (in_word)
		{
		    if (argc >= maxargs)
		    {
			char **oldargv = argv;
			maxargs *= 2;
			argv = AllocRemember(&AL->ArgMemList,
					     maxargs * sizeof(char *), 0);
			if (!argv)
			{
			    FreeArgs(AL);
			    return 0;
			}

			memcpy(argv, oldargv, maxargs/2*sizeof(char *));
			FreeRemember(oldargv);
		    }
		    *currarg = '\0';
		    if (!(argv[argc] = AllocRemember(&AL->ArgMemList,
						     currarglen + 1, 0)))
		    {
			FreeArgs(AL);
			return 0;
		    }
		    memcpy(argv[argc], argtmp, currarglen);
		    argv[argc++][currarglen] = '\0';
		    in_word = FALSE;
		}
	    }
	    else
	    {
		if (!in_word)
		{
		    currarg = argtmp;
		    in_word = TRUE;
		    currarglen = 0;
		}

		switch (*args)
		{
		case '"':
		    in_par = 1 - in_par;
		    break;
		case '\\':
		    if (!(*(args+1)))  /* If next char == NULL... */
			break;         /*   ...do nothing */
		    args++;            /* else add read ptr and FALL THRU */
		default:
		    if (currarglen < MAX_ARG_LEN)
		    {
			*currarg++ = *args;
			currarglen++;
		    }
		    break;
		}
	    }
	} while (*args++);

	AL->argc = argc;
	AL->argv = argv;
    }

    return AL;
}


void _FreeArgs(struct ArgList *AL)
{
    FreeRememberAll(&AL->ArgMemList);
    FreeMem(AL, sizeof(struct ArgList));
}
