#include "lists.i"
.text
|---------------------------------------------------------------------------
|  Insert( list, node, listNode )
|          A0    A1    A2
|---------------------------------------------------------------------------
Insert:
	movel	A2,D0		| Check if node to insert after is zero.
	beq	AddHead		| If so, do an AddHead instead.
	movel	A2@,D0		| See if list node indicates end of list.
	beq	EndOfList	| Branch if so.

 | If we get this far, A2 must point to a list node.
	movel	D0,A0
	moveml	D0/A2,A1@	| Set forward and back pointers of new node.
	movel	A1,A0@(4)	| Set back pointer of next node.
	movel	A1,A2@		| Set forward pointer of previous node.
	rts

 | The "Successor" field of the listNode was zero, i.e. listNode
 | is really the end of list marker.  In this case, insert just
 | before listNode, i.e. at the end of the list (undocumented).

EndOfList:
	movel	A2,A1@		| Set forward pointer of new node to end.
	movel	A2@(4),A0	| Get old TailPred pointer
	movel	A0,A1@(4)	| Link previous end node to new node.
	movel	A1,A2@(4)	| Set the TailPred pointer to the new node.
	movel	A1,A0@		| Link this node to the previous end node.
	rts

|---------------------------------------------------------------------------
|  AddHead( list, node )
|           A0    A1
|---------------------------------------------------------------------------
AddHead:
	movel	A0@,D0		| Get old pointer to first node.
	movel	A1,A0@		| Make the first node the new node.
	moveml	D0/A0,A1@	| Set forward and back pointers in new node.
	movel	D0,A0
	movel	A1,A0@(4)	| Fix back pointer in previous first node.
	rts

|---------------------------------------------------------------------------
|  AddTail( list, node )
|           A0    A1
|---------------------------------------------------------------------------
AddTail:
	lea	A0@(4),A0	| Point A0 to list's "Tail" field.
	movel	A0@(4),D0	| Save old value of TailPred field.
	movel	A1,A0@(4)	| Make the new node the last node.
	movel	A0,A1@		| Set new node's back pointer to list header.
	movel	D0,A1@(4)	| Set new node's back pointer.
	movel	D0,A0
	movel	A1,A0@		| Link new node to previous last node.
	rts

|---------------------------------------------------------------------------
|  Remove( node )
|          A1
|---------------------------------------------------------------------------
Remove:
	movel	A1@,A0		| Get pointer to successor node.
	movel	A1@(4),A1	| Get pointer to predecessor node.
	movel	A0,A1@		| Point predecessor to successor.
	movel	A1,A0@(4)	| Point successor to predecessor.
	rts

|---------------------------------------------------------------------------
|  node = RemHead( list )
|  D0              A0
|---------------------------------------------------------------------------
RemHead:
	movel	A0@,A1		| Get list's head pointer.
	movel	A1@,D0		| Get first node's successor.
	beq	EndRemHead	| If zero, list is empty, so return zero.

	movel	D0,A0@		| Point head pointer past removed node.
	exg	A1,D0
	movel	A0,A1@(4)	| Point new head node back at list header.
EndRemHead:
	rts

|---------------------------------------------------------------------------
|  node = RemTail( list )
|  D0              A0
|---------------------------------------------------------------------------
RemTail:
	movel	A0@(8),A1	| Get list's TailPred pointer.
	movel	A1@(4),D0	| Check if the last node has a predecessor.
	beq	EndRemTail	| If not, the list is empty, so return zero.
	movel	D0,A0@(8)	| Point TailPred pointer at second-last node.
	exg	A1,D0
	movel	A0,A1@		| Point second-last node at the list header.
	addql	#4,A1@		| Increment to point at "Tail" field.
EndRemTail:
	rts

|---------------------------------------------------------------------------
|  Enqueue( list, node )
|           A0    A1
|---------------------------------------------------------------------------
Enqueue:
	moveb	A1@(ln_Pri),D1	| Get the priority field of the node.
	movel	A0@,D0		| Get the head pointer of the queue.
        | Loop:  Keep skipping nodes until the end of the list is
        | reached or the next node has a lower priority.
EnqueueLoop:
	movel	D0,A0		| Point at the current node.
	movel	A0@,D0		| Check if there is a successor.
	beq	EndEnqueue	| If not, we are at the end of the list.
	cmpb	A0@(ln_Pri),D1	| Compare priorities.
	ble	EnqueueLoop	| Loop until priority is lower than new node's.
        | A0 now points at the node to insert before, or at the end
        | of the list (i.e. the list header plus 4).
EndEnqueue:
	movel	A0@(4),D0	| Get old value of back pointer
				| (or TailPred pointer of list header).
	movel	A1,A0@(4)	| Change it to point to the new node.
	movel	A0,A1@		| Set new node's forward pointer.
	movel	D0,A1@(4)	| Set new node's back pointer to last node
				| skipped, or list header.
	movel	D0,A0
	movel	A1,A0@		| Set list's head field, or forward pointer
				| of previous node, to new node.
	rts


|---------------------------------------------------------------------------
|  NewList( list )
|           A0
|---------------------------------------------------------------------------
NewList:
	movel	A0,A0@
	addql	#4,A0@
	clrl	A0@(lh_Tail)
	movel	A0,A0@(lh_TailPred)
	rts



|---------------------------------------------------------------------------
|	node=FindName(start, name)
|	D0		A0   A1
|---------------------------------------------------------------------------

FindName:
	movel	A2,SP@-
	movel	A0,A2		| Point to list header or first node.
	movel	A1,D1		| Save address of match string in D1.
	movel	A2@,D0
	beqs	FindEnd		| Exit if list is empty.

| Main loop:	Check if node pointed to by D0 has the same name
| as the string pointed to by D1.
FindMain:
	movel	D0,A2		| Get the address of the node to check.
	movel	A2@,D0
	beqs	FindEnd		| Exit if this is the end of the list.

	movel	A2@(ln_Name),A0	| Point A0 at name of this node.
	cmpw	#0,A0		| tstl	A0
	beqs	FindMain	| ln_Name==NULL ?
	movel	D1,A1		| Point A1 at match string.

| Inner loop:	Compare the strings.
FindLoop:
	cmpmb	A0@+,A1@+	| Compare one character.
	bnes	FindMain	| Skip to next node if not equal.
	tstb	A0@(-1)		| Was this the terminating zero?
	bnes	FindLoop	| Continue comparing if not.

	movel	A2,D0		| Got a match - point D0 to the node.

FindEnd:
	movel	D1,A1		| Restore registers and exit.
	movel	SP@+,A2
	rts

|---------------------------------------------------------------------------
|	node=HeadNode(list)
|	D0	       A0
|---------------------------------------------------------------------------
|---------------------------------------------------------------------------
|	node=NextNode(node)
|	D0	       A0
|---------------------------------------------------------------------------
|	(the code for these two functions happens to be the same)
HeadNode:
NextNode:
	movel	A0@/*(lh_Head)*/,A0
	tstl	A0@/*(ln_Next)*/
	bne	ValidNode
	bra	InvalidNode

|---------------------------------------------------------------------------
|	node=TailNode(list)
|	D0	       A0
|---------------------------------------------------------------------------
|---------------------------------------------------------------------------
|	node=PrevNode(node)
|	D0	       A0
|---------------------------------------------------------------------------
|	(weird optimizations :-) )
TailNode:
	lea	A0@(4),A0
PrevNode:
	movel	A0@(ln_Prev),A0
	tstl	A0@(ln_Prev)
	bne	ValidNode
InvalidNode:
	subl	A0,A0
ValidNode:
	movel	A0,D0
	rts

