#include <string.h>

#include <exec/tasks.h>
#include <dos/dos.h>
#include <dos/dostags.h>
#include <exec/exec_protos.h>

#define N_UNDF  0x0             /* undefined */
#define N_ABS   0x2             /* absolute */
#define N_TEXT  0x4             /* text */
#define N_DATA  0x6             /* data */
#define N_BSS   0x8             /* bss */
#define N_COMM  0x12            /* common (internal to ld) */
#define N_FN    0x1e            /* file name symbol */

#define N_EXT   01              /* external bit, or'ed in */
#define N_TYPE  0x1e            /* mask for all the type bits */

struct reloc_info_68k
{
    long  r_address;    /* address which is relocated */
    unsigned int
    r_symbolnum:24,     /* local symbol ordinal */
    r_pcrel:1,          /* was relocated pc relative already */
    r_length:2,         /* 0=byte, 1=word, 2=long */
    r_extern:1,         /* does not include value of sym referenced */
    r_baserel:1,        /* linkage table relative */
    r_jmptable:1,       /* pc-relative to jump table */
    r_relative:1,       /* relative relocation */
        :1;
};

static int Relocate(UBYTE *segment, /* start of current relocated seg */
		    struct reloc_info_68k *reloc, /* start of current reloc */
		    ULONG relocnum, /* number of relocs */
		    UBYTE *text,
		    UBYTE *data,
		    UBYTE *bss);

static void ProcExpunge(void);


struct Process *CreateNewProcTags( ULONG tag1type, ... )
{
    struct Process *child = AllocMem(sizeof(struct Process), MEMF_CLEAR);
    ULONG flags = 0, entry = 0;
    ULONG *arg = &tag1type;
    struct SegList *segl = NULL;
    APTR input = NULL, output = NULL, error = NULL, currentdir = NULL;
    int closeinput = TRUE, closeoutput = TRUE;
    int closeerror = TRUE, freeseglist = TRUE, copyvars = TRUE;
    ULONG stacksize = 4000;
    char *procname = "New Process";
    LONG procpri = FindTask(NULL)->tc_Node.ln_Pri;
    char *argString = NULL;

    if(!child)
	return NULL;

    while(*arg != TAG_DONE)
    {
	switch(*arg++)
	{
	case NP_Seglist:
	    segl = (struct SegList *)(*arg++);
	    break;
	case NP_FreeSeglist:
	    freeseglist = *arg++;
	    break;
	case NP_Entry:
	    entry = (ULONG)(*arg++);
	    break;
	case NP_Input:
	    input = *arg++;
	    break;
	case NP_Output:
	    output = *arg++;
	    break;
	case NP_Error:
	    error = *arg++;
	    break;
	case NP_CloseInput:
	    closeinput = *arg++;
	    break;
	case NP_CloseOutput:
	    closeoutput = *arg++;
	    break;
	case NP_CloseError:
	    closeerror = *arg++;
	    break;
	case NP_CurrentDir:
	    currentdir = (APTR)(*arg++);
	    break;
	case NP_StackSize:
	    stacksize = *arg++;
	    break;
	case NP_Name:
	    procname = (char *)(*arg++);
	    break;
	case NP_Priority:
	    procpri = (LONG)(*arg++);
	    break;
	case NP_Arguments:
	    argString = (char *)(*arg++);
	    break;
	case NP_CopyVars:
	    copyvars = *arg++;
	    break;
	default:
	    arg++;
	}
    }

    stacksize &= ~1; /* Round to the lower even length */
    child->pr_Task.tc_StackSize = stacksize;
    if(!(child->pr_Task.tc_Stack = AllocMem(stacksize, 0)))
    {
	FreeMem(child, sizeof(struct Process));
	return NULL;
    }
    child->pr_StackTop = child->pr_Task.tc_Stack + (stacksize/2);
    child->pr_StackBottom = child->pr_Task.tc_Stack;
    if(!(child->pr_Task.tc_Node.ln_Name = (char *)AllocMem(strlen(procname)+1, 0)))
    {
	FreeMem(child, sizeof(struct Process));
	FreeMem(child->pr_Task.tc_Stack, stacksize);
	return NULL;
    }
    strcpy(child->pr_Task.tc_Node.ln_Name, procname);

    /* Process-specific initialization */
    if(argString)
    {
	if((child->pr_Arguments = (char *)AllocMem(strlen(argString)+1, 0)))
	{
	    strcpy(child->pr_Arguments, argString);
	    flags |= PRF_FREEARGS;
	}
    }
    NewList(&child->pr_LocalVars);

    /* TODO: copy local vars etc. Also free them at expunge.. */

    if(FindTask(NULL)->tc_Node.ln_Type==NT_PROCESS)
    {
	if(!input)
	{
	    input = Open("NIL:", MODE_OLDFILE);
	    closeinput = TRUE;
	}
	if(!output)
	{
	    output = Open("NIL:", MODE_OLDFILE);
	    closeoutput = TRUE;
	}
	if(!error)
	{
	    error = Open("NIL:", MODE_OLDFILE);
	    closeerror = TRUE;
	}
	if(!currentdir &&
	   ((struct Process *)FindTask(NULL))->pr_CurrentDir)
	    currentdir = Lock("", ACCESS_READ);
    }

    if(input && closeinput)
	flags |= PRF_CLOSEINPUT;
    if(output && closeoutput)
	flags |= PRF_CLOSEOUTPUT;
    if(error && closeerror)
	flags |= PRF_CLOSEERROR;
    if(freeseglist && !entry)
	flags |= PRF_FREESEGLIST;
    if(currentdir)
    {
	child->pr_CurrentDir = currentdir;
	flags |= PRF_FREECURRDIR;
    }
    NewList(&child->pr_Path);

    child->pr_Flags = flags;
    child->pr_CIS = input;
    child->pr_COS = output;
    child->pr_CES = error;
    child->pr_CurrentDir = currentdir;
    child->pr_SegList = (APTR)segl;
    Forbid();
    child->pr_Task.tc_Node.ln_Pri = procpri;

    if(entry)
    {
	AddTask(&child->pr_Task,
		entry,
		(ULONG)ProcExpunge);
    }
    else
    {
	AddTask(&child->pr_Task,
		(ULONG)segl+sizeof(struct SegList),
		(ULONG)ProcExpunge);
    }
    /* AddTask initializes the ln_Type field to NT_TASK ! */
    child->pr_Task.tc_Node.ln_Type = NT_PROCESS;
    Permit();
    return child;
}


LONG SystemTags( const char *command, ULONG tag1type, ... )
{
    /* Start a shell to interpret the command line */
    /* If command==NULL, will start an interactive shell */
    /* SYS_Async=TRUE spawns the command and returns */

    return -1;
}


struct Process *CreateProc(const char *name,
			   LONG pri,
			   struct SegList *segList,
			   ULONG stackSize,
			   const char *argString)
{
    return CreateNewProcTags(NP_Seglist, segList,
			     NP_StackSize, stackSize,
			     NP_Name, name,
			     NP_Priority, pri,
			     NP_Arguments, argString,
			     TAG_DONE);
}


/* Size of the process structure for exec _Expunge */
const ULONG __process_struct_size = sizeof(struct Process);
extern void Switch(void);
static void ProcExpunge(void)
{
    struct Process *proc = (struct Process *)FindTask(NULL);
    ULONG flags = proc->pr_Flags;
    struct List *varlist =
	&((struct Process *)FindTask(NULL))->pr_LocalVars;
    struct LocalVar *var;
    struct AssignList *al;

    /* Free local variables */
    while((var = (struct LocalVar *)RemHead(varlist)))
    {
	Remove((struct Node *)var);
	FreeMem(var->lv_Value, var->lv_Len);
	FreeMem(var, sizeof(struct LocalVar)+strlen(var->lv_Node.ln_Name)+1);
    }

    if((flags & PRF_FREESEGLIST) && proc->pr_SegList)
	UnLoadSeg((struct SegList *)proc->pr_SegList);

    if((flags & PRF_FREECURRDIR) && proc->pr_CurrentDir)
	UnLock(proc->pr_CurrentDir);

    if((flags & PRF_CLOSEINPUT) && proc->pr_CIS)
	Close(proc->pr_CIS);

    if((flags & PRF_CLOSEOUTPUT) && proc->pr_COS)
	Close(proc->pr_COS);

    if((flags & PRF_CLOSEERROR) && proc->pr_CES)
	Close(proc->pr_CES);

    if((proc->pr_Flags & PRF_FREEARGS) && proc->pr_Arguments)
	FreeMem(proc->pr_Arguments, strlen(proc->pr_Arguments)+1);


    while((al = (struct AssignList *)RemHead(&proc->pr_Path)))
    {
	UnLock(al->al_Lock);
	FreeMem(al, sizeof(struct AssignList));
    }

    if(proc->pr_Task.tc_Node.ln_Name)
	FreeMem(proc->pr_Task.tc_Node.ln_Name,
		strlen(proc->pr_Task.tc_Node.ln_Name)+1);

    FreeRememberAll(&proc->pr_Task.tc_MemList);
    /* SetTaskPri((struct Task *)proc,-128);*/
    proc->pr_Task.tc_State = TS_INVALID;
    Switch();
    /* Never run again */
}

extern LONG PrimitiveRunCommand(UWORD *stacktop, ULONG jumpaddress);
LONG RunCommand(struct SegList *segList, ULONG stacksize,
		const char *args, ULONG arglen)
{
    struct Process *proc = (struct Process *)FindTask(NULL);
    char *oldargs = proc->pr_Arguments;
    UWORD *newstack = (UWORD *)AllocMem(2*stacksize, 0);
    UWORD *stacktop = proc->pr_StackTop, *stackbottom = proc->pr_StackBottom;
    UWORD *tc_Stack = proc->pr_Task.tc_Stack;
    ULONG tc_StackSize = proc->pr_Task.tc_StackSize;
    LONG error;

    if(!newstack)
	return -1;

    proc->pr_Arguments = (char *)args;

    proc->pr_StackTop = newstack + stacksize; /* stacksize in words, UWORD * */
    proc->pr_StackBottom = newstack;
    proc->pr_Task.tc_Stack = newstack;        /* to make it simpler for top */
    proc->pr_Task.tc_StackSize = 2*stacksize; /* size in bytes */

    error = PrimitiveRunCommand(newstack+stacksize-4, (ULONG)(segList + 1));

    proc->pr_Task.tc_Stack = tc_Stack;
    proc->pr_Task.tc_StackSize = tc_StackSize;
    proc->pr_StackTop = stacktop;
    proc->pr_StackBottom = stackbottom;

    proc->pr_Arguments = oldargs;

    FreeMem((char *)newstack, 2*stacksize);
    return error;
}

struct SegList *LoadSeg(const char *name)
{
    struct Executable exe;
    UBYTE *text=NULL, *data=NULL, *bss=NULL, *tmp;
    struct SegList *seg = NULL;
    APTR file;
    LONG error = 0;

    if(!(file = Open((char *)name, MODE_OLDFILE)))
    {
	error = ERROR_OBJECT_NOT_FOUND;
	goto errexit;
    }

    if((Read(file, (char *)&exe, sizeof(exe)) != sizeof(struct Executable)) ||
       !exe.exe_Text ||
       (exe.exe_Magic&0x7fffffff)!=0x44010107)
    {
	error = ERROR_OBJECT_WRONG_TYPE;
	goto errexit;
    }
    if(!(text = (UBYTE *)AllocMem(exe.exe_Text+sizeof(struct SegList),0)))
    {
	error = ERROR_NO_FREE_STORE;
	goto errexit;
    }

    seg = (struct SegList *)text;
    seg->sl_Length = exe.exe_Text + sizeof(struct SegList);
    seg->sl_Next = NULL;
    
    if(exe.exe_Data)
    {
	data = (UBYTE *)AllocMem(exe.exe_Data+sizeof(struct SegList),0);
	if(!data)
	{
	    error = ERROR_NO_FREE_STORE;
	    goto errexit;
	}
	seg->sl_Next = (struct SegList *)data;
	seg->sl_Next->sl_Next = NULL;
	seg->sl_Next->sl_Length = exe.exe_Data + sizeof(struct SegList);
	
	if(exe.exe_BSS)
	{
	    bss = (UBYTE *)AllocMem(exe.exe_BSS + sizeof(struct SegList),
				    MEMF_CLEAR);
	    if(!bss)
	    {
		error = ERROR_NO_FREE_STORE;
		goto errexit;
	    }
	    ((struct SegList *)bss)->sl_Length =
		exe.exe_BSS + sizeof(struct SegList);
	    ((struct SegList *)bss)->sl_Next = NULL;
	    ((struct SegList *)data)->sl_Next = (struct SegList *)bss;
	}
    }
    else
    {
	/* No data hunk */
	if(exe.exe_BSS)
	{
	    bss = (UBYTE *)AllocMem(exe.exe_BSS + sizeof(struct SegList),
				    MEMF_CLEAR);
	    if(!bss)
	    {
		error = ERROR_NO_FREE_STORE;
		goto errexit;
	    }
	    seg->sl_Next = (struct SegList *)bss;
	    seg->sl_Next->sl_Next = NULL;
	    seg->sl_Next->sl_Length = exe.exe_BSS + sizeof(struct SegList);
	}
    }
    /*
       OK, memory is allocated for all of the chunks, now just
       copy code and initialized data into place.
       */
    text += sizeof(struct SegList); /* Point to the actual code area */
    data += sizeof(struct SegList); /* Point to the actual data area */
    bss  += sizeof(struct SegList); /* Point to the actual bss area */
    
    if(Read(file, (char *)text, exe.exe_Text) != exe.exe_Text)
    {
	error = ERROR_OBJECT_WRONG_TYPE;
	goto errexit;
    }
    
    if(exe.exe_Data)
    {
	if(Read(file, (char *)data, exe.exe_Data) != exe.exe_Data)
	{
	    error = ERROR_OBJECT_WRONG_TYPE;
	    goto errexit;
	}
    }
    if(exe.exe_TRSize)
    {
	if(!(tmp = (UBYTE *)AllocMem(exe.exe_TRSize, 0)))
	{
	    error = ERROR_NO_FREE_STORE;
	    goto errexit;
	}
	if(Read(file, (char *)tmp, exe.exe_TRSize) != exe.exe_TRSize)
	{
	    FreeMem(tmp, exe.exe_TRSize);
	    error = ERROR_OBJECT_WRONG_TYPE;
	    goto errexit;
	}
	if(Relocate(text, (struct reloc_info_68k *)tmp,
		    exe.exe_TRSize/sizeof(struct reloc_info_68k),
		    text, data, bss))
	{
	    FreeMem(tmp, exe.exe_TRSize);
	    error = ERROR_OBJECT_WRONG_TYPE;
	    goto errexit;
	}
	
	FreeMem(tmp, exe.exe_TRSize);
    }
    if(exe.exe_DRSize)
    {
	if(!(tmp = (UBYTE *)AllocMem(exe.exe_DRSize, 0)))
	{
	    error = ERROR_NO_FREE_STORE;
	    goto errexit;
	}
	if(Read(file, (char *)tmp, exe.exe_DRSize) != exe.exe_DRSize)
	{
	    FreeMem(tmp, exe.exe_DRSize);
	    error = ERROR_OBJECT_WRONG_TYPE;
	    goto errexit;
	}
	if(Relocate(data, (struct reloc_info_68k *)tmp,
		    exe.exe_DRSize/sizeof(struct reloc_info_68k),
		    text, data, bss))
	{
	    FreeMem(tmp, exe.exe_DRSize);
	    error = ERROR_OBJECT_WRONG_TYPE;
	    goto errexit;
	}
	FreeMem(tmp, exe.exe_DRSize);
    }
    Close(file);
    return seg;

errexit:
    if(file)
	Close(file);
    if(seg)
	UnLoadSeg(seg);
    SetIoErr(error);
    return NULL;
}

struct SegList *ROMLoadSeg(const char *name)
{
    /* TODO: list locking and finally moving from exec to dos. */
    struct Program *prg = (struct Program *)FindProgram((char *)name);
    struct Executable *exe;
    UBYTE *text=NULL, *data=NULL, *bss=NULL, *tmp;
    struct SegList *seg = NULL;

    if(!prg)
	return NULL;
    exe = prg->pm_Exec;

    /* Program found ? Any code to execute ? Valid executable ? */
    if(!exe || !exe->exe_Text || (exe->exe_Magic&0xffff)!=0x0107)
	return NULL;
    if(!(text = (UBYTE *)AllocMem(exe->exe_Text + sizeof(struct SegList), 0)))
	return NULL;

    seg = (struct SegList *)text;
    seg->sl_Length = exe->exe_Text + sizeof(struct SegList);
    seg->sl_Next = NULL;
    if(exe->exe_Data)
    {
    	if(!(data = (UBYTE *)AllocMem(exe->exe_Data + sizeof(struct SegList), 0)))
    	{
	    UnLoadSeg(seg);
	    return NULL;
    	}
    	((struct SegList *)data)->sl_Length = exe->exe_Data + sizeof(struct SegList);
	((struct SegList *)data)->sl_Next = NULL;
	seg->sl_Next = (struct SegList *)data;

	if(exe->exe_BSS)
	{
	    if(!(bss = (UBYTE *)AllocMem(exe->exe_BSS + sizeof(struct SegList), MEMF_CLEAR)))
	    {
		UnLoadSeg(seg);
		return NULL;
	    }
	    ((struct SegList *)bss)->sl_Length = exe->exe_BSS + sizeof(struct SegList);
	    ((struct SegList *)bss)->sl_Next = NULL;
	    ((struct SegList *)data)->sl_Next = (struct SegList *)bss;
	}
    }
    else
    {
	/* No data hunk */
	if(exe->exe_BSS)
	{
	    if(!(bss = (UBYTE *)AllocMem(exe->exe_BSS + sizeof(struct SegList), MEMF_CLEAR)))
	    {
		UnLoadSeg(seg);
		return NULL;
	    }
	    ((struct SegList *)bss)->sl_Length = exe->exe_BSS + sizeof(struct SegList);
	    ((struct SegList *)bss)->sl_Next = NULL;
	    ((struct SegList *)text)->sl_Next = (struct SegList *)bss;
	}
    }
    /*
       OK, memory is allocated for all of the chunks, now just
       copy code and initialized data into place.
     */

    text += sizeof(struct SegList); /* Point to the actual code area */
    data += sizeof(struct SegList); /* Point to the actual data area */
    bss  += sizeof(struct SegList); /* Point to the actual bss area */

    tmp = ((UBYTE *)exe) + sizeof(struct Executable); /* Start of code */
    memcpy(text, tmp, exe->exe_Text);
    tmp += exe->exe_Text;                /* Start of data */

    if(exe->exe_Data)
    {
	memcpy(data, tmp, exe->exe_Data);
	tmp += exe->exe_Data;                /* Start of code reloc */
    }
    if(Relocate(text, (struct reloc_info_68k *)tmp,
		exe->exe_TRSize/sizeof(struct reloc_info_68k),
		text, data, bss))
    {
	UnLoadSeg(seg);
	return NULL;
    }
    tmp += exe->exe_TRSize;          /* Start of data reloc */
    if(exe->exe_Data && exe->exe_DRSize)
    {
	if(Relocate(data, (struct reloc_info_68k *)tmp,
		    exe->exe_DRSize/sizeof(struct reloc_info_68k),
		    text, data, bss))
	{
	    UnLoadSeg(seg);
	    return NULL;
	}
    }
    return seg;
}



LONG UnLoadSeg(struct SegList *seg)
{
    struct SegList *next;

    while(seg)
    {
	next = seg->sl_Next;
	FreeMem(seg, seg->sl_Length);
	seg = next;
    }
    return 0;
}


static int Relocate(UBYTE *segment, /* start of current relocated seg */
		    struct reloc_info_68k *reloc,   /* start of current reloc info */
		    ULONG relocnum,/* number of relocs */
		    UBYTE *text,
		    UBYTE *data,
		    UBYTE *bss)
{
    ULONG curtype;
    UBYTE flags;

    while(relocnum--)
    {
	if(reloc->r_extern) /* Unresolved reference! */
	{
	    return -1;
	}
	if(!reloc->r_pcrel)
	{
	    flags = (reloc->r_symbolnum & N_TYPE);
	    switch(flags)
	    {
	    case N_ABS:
		curtype = 0;
		break;
	    case N_TEXT:
		curtype = (ULONG)text;
		break;
	    case N_DATA:
		curtype = (ULONG)data;
		break;
	    case N_BSS:
		curtype = (ULONG)bss;
		break;
	    default:
	        return -1;
	    }
	    switch(reloc->r_length)   /* 0=byte, 1=word, 2=long */
	    {
	    case 0:
		*(UBYTE *)((ULONG)segment + reloc->r_address) += curtype;
		break;
	    case 1:
		*(UWORD *)((ULONG)segment + reloc->r_address) += curtype;
		break;
	    case 2:
		*(ULONG *)((ULONG)segment + reloc->r_address) += curtype;
		break;
	    }
	}
	reloc++;
    }
    return 0;
}
