#include "execbase.i"
#include "io.i"
#include "system.i"

/* WARNING!!:
	L-labels should be used with care. If you have a reference
	to one that does not exist, you silently get the address of the
	next to the last instruction in the object file.
 */

  /***************************/
 /* Start of bootstrap code */
/***************************/

#define realbase 0x800000
.text

	.long	BASE+0x200		| Initial SSP
	.long	reset-realbase		| Initial PC
	.long	deadend-realbase	| Bus Error
	.long	deadend-realbase	| Address Error
	.long	deadend-realbase	| Illegal Instruction
	.long	deadend-realbase	| Zero Divide
	.long	deadend-realbase	| CHK
	.long	deadend-realbase	| TRAPV
	.long	deadend-realbase	| Privilege Violation
	.long	deadend-realbase	| Trace
	.long	deadend-realbase	| Line 1010 Emulator
	.long	deadend-realbase	| Line 1111 Emulator
	.long	deadend-realbase	| 12
	.long	deadend-realbase	| 13
	.long	deadend-realbase	| 14
	.long	deadend-realbase	| Uninitialized Interrupt Vector
	.long	deadend-realbase	| 16
	.long	deadend-realbase	| 17
	.long	deadend-realbase	| 18
	.long	deadend-realbase	| 19
	.long	deadend-realbase	| 20
	.long	deadend-realbase	| 21
	.long	deadend-realbase	| 22
	.long	deadend-realbase	| 23
	.long	spurious-realbase	| Spurious Interrupt

deadend:
	stop	#0x2000		| Wait for an interrupt
	rte
spurious:
	rte

	.even
	| This routine should be less or equal to 6 longwords
resetstart:
	| Move EPROM + Flash to somewhere else and unveil RAM beneath it..
	| BR0 : FC=000 Base=10000000000 ($800000) RW=0 EN=1
	| OR0 : DTACK=001 Mask=11000000000 MRW=0 CFC=0 (150ns EPROM/Flash)
	movel	#0x10013800,BR0	| CS0 init (4M) (BR0 and OR0 written)
	jmp	continuereset

reset:				| This part is executed in supervisor state
	movew	#0x2700,SR
	movew	#BASE/4096,BAR	| Init BAR - On-Chip RAM mapped to $650000

	| SCR:	ERRE=0,EMWS=1,ADCE=0,LPEN=1,LPdivider=2 (default->10MHz)
	movel	#0x00040020,SCR

	| Reconfigure CS0 (EPROM in Supervisor Data Space)
	| BR0 : FC=000 Base=00000000000 ($000000) RW=0 EN=1
	| OR0 : DTACK=010 Mask=11111000000 MRW=0 CFC=0 (200ns EPROM)
	movel	#0x00015f00,BR0	| CS0 init (256k) (BR0 and OR0 written)

	| BR3 : FC=000 Base=11000000000 ($c00000) RW=0 EN=1
	| OR3 : DTACK=100 Mask=11000000000 MRW=0 CFC=0 (300ns ?)
	movel	#0x18019800,BR3	| CS3 init (4M) (BR3 and OR3 written)
	| BR2 : FC=000 Base=00000000000 ($000000) RW=0 EN=1
	| OR2 : DTACK=011 Mask=10000000000 MRW=0 CFC=0 (250ns DRAM :-()
	movel	#0x00017000,BR2	| CS2 init (8M) (BR2 and OR2 written)
	| BR1 : FC=000 Base=01100000000 ($600000) RW=0 EN=1
	| OR1 : DTACK=110 Mask=11100000000 MRW=0 CFC=0 (400ns ?)
	movel	#0x0c01dc00,BR1	| CS1 init (2M) (BR1 and OR1 written)

	| Initialize Ethernet controller
	| BMODE = 1 (Big Endian) DW = 0 (16-bit)
	| DCR [00000000 00000000]
	movew	#0x0080,E_CR	| Disable SONIC (RST = 1)
	movew	#0x8000,E_DCR	| EXBUS=1
	movew	#0,E_IMR	| Interrupts disabled
	movew	#0x0000,E_DCR2	| LEDs on for bus master cycles

	| Initialize other I/O pins
	movew	#0x0040,PBCNT	| /TOUT2 as dedicated pin
	movew	#0,PBDAT	| TODO:	The REAL values!!
	movew	#0x0e41,PBDDR	| PB2-5,7-8 as inputs, PB1,6,9-11 as outputs

	movew	#0x00c0,SIMODE	| Normal operation - no multiplexing
	movew	#0x0033,PACNT	| SCC2 TxD,RxD,CTS,RTS as dedicated
	movew	#0,PADAT	| TODO:	The REAL values!!
	movew	#0x874c,PADDR

	moveb	#0x88,I8255+2*3+1	| Configure A,B,C(lo) OUT, C(hi) IN

	| Init memory refreshes.. one refresh per 312 clocks at 20MHz
	| We skip 2k+2, so it doesn't matter if /RAS latches LSb's or MSb's.
	| There are 3 banks, 1M*16bit each. They are refreshed simultaneously.
	movel	#0x00000000,BASE+0x570	| Refresh starts at 0
	movel	#0x08020400,BASE+0x574	| skip 2050, 1024 rows
	movew	#0,BASE+0x57c		| Init temporary count
	movew	#312,TRR2		| Timer reference value
	movew	#0,IMR			| IMR = 0, disable internal interrupts
	| Start timer2 PS=0x00,CE=00,OM=1,ORI=0,FRR=1,ICLK=01,RST=1
	movew	#0x000b,TMR2		| Toggle mode
	| Memory refreshes now start
	| SCR:	ERRE=1,EMWS=1,ADCE=0,HWDEN=1,LPEN=1,LPdivider=2 (->10MHz)
	movel	#0x00440820,SCR
	movew	#0x8340,GIMR	| On-chip vectors start at 64

	| move [resetstart .. resetend) to Dual-port ram, jump there
	lea	resetstart-realbase,A0
	lea	BASE,A1
	movel	A0@+,A1@+
	movel	A0@+,A1@+
	movel	A0@+,A1@+
	movel	A0@+,A1@+
	movel	A0@+,A1@+	| One extra longword just for safety
	jmp 	BASE

  /*************************/
 /* End of bootstrap code */
/*************************/
	
/* #define TEST_PB8INT */

/* These macros do not use any registers, thus they are macros :-) */
#define FORBID		addqw #1,A5@(TDNestCnt)
#define PERMIT		subqw #1,A5@(TDNestCnt)
#define WDOG_TOGGLE	eorw  #master_WDOG,PADAT

#define QUANTUM     10    /* How many 10ms slices tasks may use */

#define MEMSTART    4096
/* Exception jump vectors upto 0x0c00 (currently 88 vectors -> 0x6c0) */
/* Global variables are linked to 0x0c00, size should be <0x400 */
#define SSTACKSIZE  4096
#define USTACKSIZE  4096

/* TODO: Move the supervisor stack to on-chip ram for quicker access. */
/*       May need to change serial.device accordingly. */

.globl	_strlen

	.text
	.align

#define MEM_PAGES	0xc000	/* Must fit into a word!! */
#define MEM_PAGESIZE	0x80	/* MEM_PAGES*MEM_PAGESIZE == 0x600000 */

continuereset:
	| SIMM start 0x000000
	subal	A3,A3
	| Test the memory speed - NOT 100% reliable..
	movel	#0x00011000,D1	| Mask + Enable, 0 wait states
	moveql	#5,D0		| 6 different wait state settings to test
	movel	#0xaa5555aa,D4
	movel	#0x55aaaa55,D5
	movel	#0x0000ffff,D6
	movel	#0xffff0000,D7
Lspeed:
	| BR2 : FC=000 Base=00000000000 ($000000) RW=0 EN=1
	| OR2 : DTACK=nnn Mask=10000000000 MRW=0 CFC=0 (100ns DRAM ?)
	movel	D1,BR2		| CS2 init (8M) (BR2 and OR2 written)
	movel	D4,A3@
	cmpl	A3@,D4
	bne	Lnextspeed
	movel	D5,A3@
	cmpl	A3@,D5
	bne	Lnextspeed
	movel	D6,A3@
	cmpl	A3@,D6
	bne	Lnextspeed
	movel	D7,A3@
	cmpl	A3@,D7
	beq	LspeedOK	| Does work
Lnextspeed:
	addw	#0x2000,D1	| Add one wait state
	dbf	D0,Lspeed
LspeedOK:
	| OK, we found out the max speed
	addw	#0x2000,D1	| Add one wait state
	movel	D1,BR2		| CS2 init (8M) (BR2 and OR2 written)

	| Then, (try to) clear the first word in each page.
	| During the next pass we first check the word:
	| 1) If it is 0, check the page for bit-errors. If OK,
	|    mark the page checked and advance to the next page.
	| 2) If it is not 0, either the word is bad, or there is
	|    no memory at all, or we simply have arrived to the
	|    second image of a smaller-than-1M SIMM.
	| Whatever the error condition, the location we reached
	| will become the top of memory and execbase, supervisor
	| stack, the first task structure and user stack will
	| be allocated from that location downwards. The rest
	| is added to the free memory list.

	movel	#MEM_PAGES-1,D0
	moveql	#0,D1
	movel	#MEM_PAGESIZE,D2
	subal	A3,A3
Lmemclr:
	movew	D1,A3@
	addal	D2,A3
	dbf	D0,Lmemclr

	movel	#MEM_PAGES-1,D0
	movel	#0xaa5555aa,D4
	movel	#0xa5a55a5a,D5
	movel	#MEM_PAGESIZE,D1
	subal	A3,A3
Lmemtest:
	| Allows SIMMs smaller than 1M to be used in the last bank in use.
	tstw	A3@		| Detect wrapped, bad, or open space
	bne	Lmemfail
	movel	D4,A3@		| Detect bad memory
	cmpl	A3@,D4
	bne	Lmemfail
	movel	D5,A3@
	cmpl	A3@,D5
	bne	Lmemfail
	movew	D0,A3@		| Store the page number
	addal	D1,A3		| Advance to the next page

	WDOG_TOGGLE	 	| Toggle the master watchdog

	dbf	D0,Lmemtest
	bra	LmemOK

Lmemfail:
	|movel	#USTACKSIZE+SSTACKSIZE+tc_SIZE+exec_SIZE+LIB_NEGSIZE,D0
	|cmpl	A3,D0

	movel	A3,D0
	bne	LmemOK

Lnomem:
	jsr	DebugInit
	lea	strNoMemory,A0
	moveql	#12,D0
	jsr	DebugWrite

	stop	#0x2000
	bra	Lnomem		| Wait until we are reset

LmemOK:
	WDOG_TOGGLE			| Toggle the master watchdog

	| MEMEND in A3
	subl	#exec_SIZE,A3		| reserve space for ExecBase
	movel	A3,A5			| ExecBase *
	subl	#LIB_NEGSIZE,A3		| ExecBase negative size
	lea	A3@(-8),SP		| Supervisor stack grows downwards
	subl	#SSTACKSIZE,A3

	jsr	DebugInit
	pea	strStackOK
	jsr	_DebugWrite
	addqw	#4,SP
	WDOG_TOGGLE			| Toggle the master watchdog

|-----------------
| ExecBase init
| A5 = Start of execbase
|-----------------
	movel	A5,AbsExecBase		| Init AbsExecBase
	| The stack SHOULD be OK now, we can jsr..
	jsr	InitFuncts		| Init traps and jump table
	WDOG_TOGGLE		 	| Toggle the master watchdog

	moveql	#0,D0
	movew	#QUANTUM,A5@(Quantum)
	movew	D0,A5@(AttnResched)
	movew	#2,A5@(Version)
	movel	D0,A5@(DispCount)
	movew	D0,A5@(TDNestCnt)
	movew	D0,A5@(IDNestCnt)

	lea	A5@(ResourceList),A0
	jsr 	A5@(LIB_NewList)
	lea	A5@(DeviceList),A0
	jsr 	A5@(LIB_NewList)
	lea	A5@(LibList),A0
	jsr 	A5@(LIB_NewList)
	lea	ExecName,A1
	moveb	#NT_LIBRARY,A5@(LibNode+lib_Node+ln_Type)
	movel	A1,A5@(LibNode+lib_Node+ln_Name)
	movew	ExecVersion,A5@(LibNode+lib_Version)
	movew	#1,A5@(LibNode+lib_OpenCnt)
	movel	A5,A1		| Node to add, A0 = list to add to
	lea	A5@(LibList),A0
	jsr	A5@(LIB_AddHead)

	lea	A5@(PortList),A0
	jsr	A5@(LIB_NewList)
	lea	A5@(TaskReady),A0
	jsr	A5@(LIB_NewList)
	lea	A5@(TaskWait),A0
	jsr	A5@(LIB_NewList)
	lea	A5@(TimerRequests),A0
	jsr	A5@(LIB_NewList)
	lea	A5@(ProgramList),A0
	jsr	A5@(LIB_NewList)
	lea	A5@(SemaphoreList),A0
	jsr	A5@(LIB_NewList)

	movel	#0x80000000,A5@(SystemTime+ds_Days)
	clrl	A5@(SystemTime+ds_Minutes)
	clrl	A5@(SystemTime+ds_Ticks)

|------------------------------------------------
| Reserve space for task stack and task structure
|------------------------------------------------
	subl	#8,A3
	movel	A3,USP			| Stack grows downwards
	subl	#USTACKSIZE-8,A3
	movel	A3,A1			| 'End' of stack
	subl	#tc_SIZE,A3

|-----------------
| Init memory list
|-----------------
	movel	#MEMSTART,D7	| Start of the RAM
	addl	#255,D7		| Align to page boundary
	andl	#0xffffff00,D7
	movel	D7,A5@(MemoryList) | Insert first memory node to Execbase
	moveal	D7,A4
	movel	A3,D6		| End of memory
	subl	D7,D6		| Get the size of the available memory
	clrl	A4@(mm_Next)	| mm_Next = NULL
	movel	D6,A4@(mm_Size)	| mm_Size = MEMEND-MEMSTART

	| TODO: if the top of memory is < 0x200000, check
	|	for more memory in banks 1 and 2

	pea	strMemoryListOK
	jsr	_DebugWrite
	addqw	#4,SP
	WDOG_TOGGLE		| Toggle the master watchdog

|-----------------
| Task init
|-----------------
	movel	A3,A4			| first struct Task *
	movel	A4,A5@(ThisTask)	| Init ExecBase->ThisTask
	movel	#USTACKSIZE,A4@(tc_StackSize)
	movel	A1,A4@(tc_Stack)
	moveql	#0,D1		| Use default FinalPC
	moveql	#0,D0		| InitialPC (but we are already running..)
	movel	A4,A1		| Task *
	moveb	#0,A1@(tc_Node+ln_Pri)
	clrw	A1@(tc_UserID)
	clrw	A1@(tc_GroupID)
	jsr	A5@(LIB_AddTask)| Add this task to system

	movew	#TS_RUN,A4@(tc_State) | It's us and we ARE running currently
	movel	A4,A1
	jsr	A5@(LIB_Remove)	| Remove the node from the Ready queue,
				| because we are currently running..
	WDOG_TOGGLE	 	| Toggle the master watchdog

	jsr	inittimer	| Start timer and install handler

	pea	strTimerOK
	jsr	_DebugWrite
	addqw	#4,SP
	WDOG_TOGGLE	 	| Toggle the master watchdog

#if 0
|	tstw	0xc001face
|	andib	#0xff,CCR
|	movew	CCR,D0
|	movew	D0,CCR
|	orib	#0,CCR
|	trapv
#endif
	movew	#0x0000,SR	| Continue running in user state
				| and enable interrupts
	jsr	__init		| Jump to the first program
	| We should never return, but if we do, the expunge routine
	| gets called and the task structure and stack is freed.
	jmp	__Expunge

	.even
| Debug serial routines for serial port 0
DebugInit:
	movew	#0x3e,SCCConf0+SCON	| 38400 bps
	movew	#0x0131,SCCConf0+SCM	| Async, T&R disabled, 8N1
	moveb	#0,SCC0+RFCR
	moveb	#0,SCC0+TFCR
	movew	#256,SCC0+MRBLR
	movew	#16,SCC0+MAX_IDL
	movew	#0x0a,SCC0+CHARACTER1
	movew	#0x0d,SCC0+CHARACTER2
	movew	#0x8000,SCC0+CHARACTER3
	movew	#0,SCC0+CHARACTER8
	moveb	#0,SCC0+RBDN
	moveb	#0x40,SCC0+TBDN
	moveb	#0xff,SCCConf0+SCCE	| Clear events
	moveb	#0,SCCConf0+SCCM	| Disable interrupts
	moveb	#0x11,CR		| Restart transmit command
Linitwait:
	btst	#0,CR
	bne	Linitwait
	movew	#0x0131+4,SCCConf0+SCM	| Enable transmitter
	rts

.globl _DebugWrite
_DebugWrite:
	movel	SP@(4),SP@-
	jsr	_strlen
	addql	#4,SP
	movel	SP@(4),A0
DebugWrite:	| Uses only BD0
	movel	A0,SCC0+0x40+BD_Address
	movew	D0,SCC0+0x40+BD_Length
	movew	#0xe000,SCC0+0x40+BD_Status	| Filled,External buf,Wrap
Debugwait:
	btst	#7,SCC0+0x40+BD_Status	| Test bit 15
	bne	Debugwait
	rts

.globl _DebugRead
_DebugRead:
	movel	SP@(4),A0
	movel	A0,SCC0+0+BD_Address
	movew	#1,SCC0+MRBLR
	orw	#8,SCCConf0+SCM
	movew	#0xe000,SCC0+0+BD_Status
Debugwait2:
	btst	#7,SCC0+0+BD_Status
	bne	Debugwait2
	movew	SCC0+0+BD_Length,D0
	rts

	.even
InterruptKludge:
	jsr	__SysInt
	rte

__FlashDummy:
	rts
	.even
jumptablestart:
	/* Remember!! 'reversed' order */
	jmp	__ReleaseSemaphore
	jmp	__ObtainSemaphoreShared
	jmp	__ObtainSemaphore
	jmp	__AttemptSemaphoreShared
	jmp	__AttemptSemaphore
	jmp	__FindSemaphore
	jmp	__RemSemaphore
	jmp	__AddSemaphore
	jmp	__InitSemaphore

	jmp	__FreeSemaphore
	jmp	__FlashDummy
	jmp	__FlashDummy
	jmp	__FlashDummy
	jmp	PrevNode
	jmp	NextNode
	jmp	TailNode
	jmp	HeadNode

	| Misc
	jmp	SendTimerReq
	jmp	SetDate
	jmp	DateStamp
	| Resources
	jmp	FreeResource
	jmp	AllocResource
	jmp	RemResource
	jmp	AddResource
	| Libraries
	jmp	SetFunction
	jmp	CloseLibrary
	jmp	OpenLibrary
	jmp	InitLibFuncs
	jmp	RemLibrary
	jmp	AddLibrary
	| Devices
	jmp	AbortIO
	jmp	WaitIO
	jmp	CheckIO
	jmp	SendIO
	jmp	DoIO
	jmp	CloseDevice
	jmp	OpenDevice
	jmp	RemDevice
	jmp	AddDevice

	| Ports/Messages
	jmp	ReplyMsg
	jmp	GetMsg
	jmp	PutMsg
	jmp	WaitPort
	jmp	FindPort
	jmp	DeletePort
	jmp	CreatePort
	jmp	RemPort
	jmp	AddPort
	| Resident programs
	jmp	FindProgram
	jmp	RemProgram
	jmp	AddProgram
	| Signals
	jmp	FreeSignal
	jmp	AllocSignal
	jmp	Signal
	jmp	Wait
	| Task-related
	jmp	SetSignal
	jmp	SetTaskPri
	jmp	FindTask
	jmp	RemTask
	jmp	AddTask

	| Memory
	jmp	FreeRememberAll
	jmp	FreeRemember
	jmp	AllocRemember
	jmp	AvailMem
	jmp	FreeMem
	jmp	AllocMem
	| Lists
	jmp	FindName
	jmp	NewList
	jmp	Enqueue
	jmp	RemTail
	jmp	RemHead
	jmp	Remove
	jmp	AddTail
	jmp	AddHead
	jmp	Insert

	| Interrupts
	jmp	Cause
	jmp	RemIntServer
	jmp	AddIntServer
	jmp	Permit
	jmp	Forbid
	jmp	Enable
	jmp	Disable
	jmp	Close
	jmp	Open
jumptableend:


	.even
InitFuncts:
	movel	#87,D0			| 88 vectors to initialize
	lea	InterruptKludge,A0
	| get "jsr __SysInt : rte", 8 words
	moveml	A0@,D1-D2

	| We use postincrement, start from Bus Error vector
	| So, we do not overwrite AbsExecBase
	lea	2*4,A0			| Vector table start
	lea	4*256+2*8,A1		| Jump table start
	lea	A5@(IntVects+2*lh_SIZE),A2	| start of IntVects

intloop:
	| A1 now points to the start of the jsr command
	| Put it to the interrupt vector
	| Check to not overwrite BAR and SCR
	cmpw	#0xf0,A0		| BAR entry
	bne	intsaved
	addl	#4*4,A0			| Skip 4 entries
	addl	#4*8,A1
	addl	#4*lh_SIZE,A2
intsaved:
	movel	A1,A0@			| Save the vector
	moveml	D1-D2,A1@		| jsr __SysInt : rte
	addql	#4,A0			| Next interrupt vector
	addql	#8,A1			| Next jump vector

	| Init the interrupt server lists
	| (can't use A5@(LIB_NewList), because it's not initialized yet)
	movel	A2,A2@
	addql	#4,A2@			| Head points to Tail
	clrl	A2@(lh_Tail)		| Tail is NULL
	movel	A2,A2@(lh_TailPred)	| TailPred points to the Listheader
	addaw	#lh_SIZE,A2		| Next IntVect
	| (adda will always operate on full 32 bits)

	dbf	D0,intloop

	| We override the defaults created earlier
	| These routines are not called through the stubs, but directly
	movel	#__TrapDisable,	4*(32+4)	| Trap #4
	movel	#__TrapEnable,	4*(32+5)	| Trap #5
	movel	#__Switcher,	4*(32+6)	| Trap #6
	movel	#__TrapCause,	4*(32+7)	| Trap #7
	movel	#__TrapSignal,	4*(32+8)	| Trap #8

	WDOG_TOGGLE

	| The jumptable is before this routine so that the assembler
	| will be able to handle two symbol references in one expression.
	| Uses word-copy so that no extra words are copied
	| ("jmp address" takes 6 bytes/3 words)
	movel	#jumptableend-jumptablestart,D0
	lsrl	#1,D0		| size/2 = number of WORDs
	subql	#1,D0
	moveal	A5,A1		| Execbase (destination end) to A1
	lea	jumptableend,A0	| Source to A0
Linitloop:
	movew	A0@-,A1@-
	dbf	D0,Linitloop

	WDOG_TOGGLE
	rts

	.even
Open:
	addqw	#1,A5@(LibNode+lib_OpenCnt)
	movel	A5,D0			| exec.library open always successful
	rts
Close:
	subqw	#1,A5@(LibNode+lib_OpenCnt)
	rts


	.even
__SysInt:
	movew	#0x2700,SR		| Disable interrupts
	link	A6,#0			| remember frame start
	moveml	D0-D7/A0-A5,SP@-	| save registers
	movel	A6@(4),D0		| Get the JSR return address
	subiw	#1030,D0		| subtract the base address + 6
	moveal	AbsExecBase,A5		| Get Execbase
	lea	A5@(IntVects),A4	| get IntVects start
	lsrl	#3,D0			| Divide by 8 to get vector number
	movel	D0,A6@(4)		| Save the vector number

	muluw	#lh_SIZE,D0		| lh_SIZE*vector number
	addal	D0,A4			| SysBase->IntVects[vector#]
	movel	A4@/*(lh_Head)*/,A4
	tstl	A4@			| (is_Node+ln_Next) Is there a node ?
	bne	SysIntLoop

	movew	A6@(8),D0		| status
	andiw	#0x2000,D0		| Supervisor ?
	bne	SysIntInt

	| Ok, no global handlers, and exception during user code
	| try tc_ExceptCode
	movel	A5@(ThisTask),A4
	beq	SysIntNone
	movel	A4@(tc_ExceptCode),A0
	beq	SysIntNone
	movel	A4@(tc_ExceptData),D0
	lea	A6@(4),A1
	movel	A1,SP@-
	movel	D0,SP@-			| Parameters pushed for C handlers
	addqw	#1,A5@(IDNestCnt)	| Increase disable count
	jsr	A0@
	subqw	#1,A5@(IDNestCnt)	| Decrease disable count
	addql	#8,SP			| Remove parameters

	bra	SysIntLeave

SysIntNone:
	| No handlers - make a nice informative message about this..
	| Maybe it is even fatal !
	moveb	#TS_EXCEPT,A4@(tc_State)

	jsr	DebugInit
	pea	strProgramError0	| Header
	jsr	_DebugWrite
	addql	#4,SP

	movel	A6@(4),D0
	asll	#2,D0
	lea	strPtrInt,A0
	movel	A0@(D0:w),SP@-		| Cause
	jsr	_DebugWrite
	addql	#4,SP

	pea	strProgramError1
	jsr	_DebugWrite
	addql	#4,SP

	movel	A4@(tc_Node+ln_Name),SP@-	| Task name
	jsr	_DebugWrite
	addql	#4,SP

	pea	newline
	jsr	_DebugWrite
	jsr	_DebugWrite
	addql	#4,SP

	movel	A6@(4),D0
	cmpl	#4,D0
	bhi	shortframe

	bsr	DispStackFrame
|	moveml	SP@+,D0-D7/A0-A5| restore registers
|	unlk	A6		| restore A6 & SP
|	addql	#4,SP		| remove vector number
|	bra	__Switcher
	bra	SysDeadEnd

shortframe:
	jsr	DispShortFrame
	bra	SysDeadEnd

SysIntInt:
	| Exception during exception and no global handler
	jsr	DebugInit
	pea	strDeadEnd
	jsr	_DebugWrite
	addql	#4,SP
	movel	A6@(4),D0
	asll	#2,D0
	lea	strPtrInt,A0
	movel	A0@(D0:w),SP@-		| Cause
	jsr	_DebugWrite
	addql	#4,SP
	pea	newline
	jsr	_DebugWrite
	addql	#4,SP

	movel	4,SP@-
	pea	strExecBase
	jsr	_DebugWriteUlong
	addql	#8,SP

	bsr	DispStackFrame

SysDeadEnd:
	bra	SysDeadEnd

SysIntLoop:
	movel	A4@(is_Data),D0
	movel	A4@(is_Code),A0
	lea	A6@(4),A1		| Stack frame start address
	movel	A1,SP@-
	movel	D0,SP@-			| Parameters pushed for C handlers

	addqw	#1,A5@(IDNestCnt)	| Increase disable count
	jsr	A0@
	subqw	#1,A5@(IDNestCnt)	| Decrease disable count
	addql	#8,SP			| Remove parameters

	movel	A4@/*(is_Node+ln_Next)*/,A4
	tstl	A4@/*(is_Node+ln_Next)*/	| Is there a node ?
	bne	SysIntLoop

SysIntLeave:
	movew	A6@(8),D0	| status
	andiw	#0x2000,D0	| Supervisor ?
	bne	__DefInt	| do not switch

	moveal	AbsExecBase,A0
	tstw	A0@(AttnResched)| SysBase->AttnResched
	beq	__DefInt	| No need to switch

	tstw	A0@(TDNestCnt)	| TDNestCnt - switch permitted ?
	beq	DoSwitch	| Change task

__DefInt:
	moveml	SP@+,D0-D7/A0-A5| restore registers
	unlk	A6		| restore A6 & SP
	addql	#4,SP		| remove vector number
	rte			| return from exception

DispStackFrame:
	| continue the informative message (PC,SP, regs)
	movel	A6@(18),SP@-	| PC
	pea	strPC
	jsr	_DebugWriteUlong
	addql	#8,SP

	moveql	#0,D0
	movew	A6@(16),D0
	movel	D0,SP@-		| status register
	pea	strSR
	jsr	_DebugWriteUword
	addql	#8,SP

	moveql	#0,D0
	movew	A6@(14),D0
	movel	D0,SP@-		| instruction register
	pea	strIR
	jsr	_DebugWriteUword
	addql	#8,SP

	movel	A6@(10),SP@-	| access address
	pea	strAccess
	jsr	_DebugWriteUlong
	addql	#8,SP

	moveql	#0,D0		| function code etc.
	movew	A6@(8),D0
	movel	D0,SP@-
	pea	strFC
	jsr	_DebugWriteUword
	addql	#8,SP
	rts
DispShortFrame:
	movel	A6@(10),SP@-	| PC
	pea	strPC
	jsr	_DebugWriteUlong
	addql	#8,SP

	moveql	#0,D0
	movew	A6@(8),D0
	movel	D0,SP@-		| status register
	pea	strSR
	jsr	_DebugWriteUword
	addql	#8,SP
	rts

	.even
DoSwitch:
	moveml	SP@+,D0-D7/A0-A5| restore registers
	unlk	A6
	addql	#4,SP		| remove vector number
	| Fall through to the switch code

__Switcher:
	movew	#0x2700,SR	| Disable all interrupts.
| Remember that no task is run when we are in exception and an interrupt
| during this will not cause another switch, so we can happily play around
| with the system tables and lists.
| Of course, interrupt routines shouldn't access system lists,
| only their own structures.

	movel	A0,SP@-			| Save A0 to the supervisor stack.
	movel	AbsExecBase,A0		| Get ExecBase
	clrw	A0@(AttnResched)	| clear AttnResched
	movel	A0@(ThisTask),A0	| Get ThisTask

	moveml	D0-D7/A1-A6,A0@(tc_Regsave)	| Store registers
	movel	SP@+,A0@(tc_Reg_a0)		| Store old A0
	| Now we have saved the registers, time for some real work

	movel	USP,A1
	movew	SP@+,A1@-	| Save the status register.
	movel	SP@+,A1@-	| Save the program counter.
	|movew	SP@+,A1@-	| Save the frame word (this means >='010)
				| Remember that the order is reversed!!
	movel	A1,A0@(tc_SPReg)	| Save the user stack pointer.

| ExecBase in A5  ThisTask in A3
	movel	AbsExecBase,A5		| ExecBase
	movel	A5@(ThisTask),A3	| ThisTask
	movew	A5@(Quantum),A3@(tc_Slice)
	| Wait() breaks Forbid()/Disable(), so remember the count for next time
	movew	A5@(TDNestCnt),A3@(tc_TDNestCnt)
	movew	A5@(IDNestCnt),A3@(tc_IDNestCnt)

	| TODO:	if the task releases the CPU voluntarily, increase priority

	cmpiw	#TS_INVALID,A3@(tc_State)
	bne	OldStillOK
| This task has made an expunge and is not to be run again, ever!
| So, we will release memory, first the stack, then the task structure
	movel	A3@(tc_StackSize),D0
	movel	A3@(tc_Stack),A0
	jsr	A5@(LIB_FreeMem)	| Free the stack
	movel	#tc_SIZE,D0		| sizeof(struct Task)
	cmpb	#NT_TASK,A3@(tc_Node+ln_Type)
	beq	notprocess
	movel	___process_struct_size,D0
notprocess:
	movel	A3,A0
	jsr	A5@(LIB_FreeMem)	| Free the task structure
	bra	SearchForTask
	
OldStillOK:
	| The task we are switching out is still valid
	lea	A5@(TaskWait),A0	| Default list to add to
	cmpiw	#TS_RUN,A3@(tc_State)	| tc_State = TS_RUN ?
	bne	AddQueue
	movew	#TS_READY,A3@(tc_State)	| ThisTask->tc_State= TS_READY
	lea 	A5@(TaskReady),A0	| We are ready...
AddQueue:
	moveal	A3,A1			| Node to add
	jsr	A5@(LIB_Enqueue)	| Put it to the right place

| From now on we don't have a valid ThisTask
	clrl	A5@(ThisTask)
SearchForTask:
	| Try to get a task from the ready queue
	lea	A5@(TaskReady),A0
	jsr	A5@(LIB_RemHead)
	tstl	D0
	bne	GotOne

	stop	#0x2000		| Wait for an interrupt - waking tasks ?
	movew	#0x2700,SR	| Disable interrupts again.
	bra	SearchForTask	| Try to find a runnable task again

| ExecBase in A5  NewTask in A3
GotOne:
	movel	D0,A3
	addql	#1,A5@(DispCount)	| Increase SysBase->DispCount
	tstw	A3@(tc_State)		| tc_State
	bne	ReLoad			| TS_INVALID ?
| We shouldn't get this far.. But just in case a ready task suddenly
| gets invalid..
| This task has made an expunge and is not to be run again, ever!
| So, we will release memory, first the stack, then the task structure
	movel	A3@(tc_StackSize),D0
	movel	A3@(tc_Stack),A0
	jsr	A5@(LIB_FreeMem)	| Free the stack
	movel	#tc_SIZE,D0		| sizeof(struct Task)
	movel	A3,A0
	jsr	A5@(LIB_FreeMem)	| Free the task structure
	bra	SearchForTask

| Pop the task's (A3) CPU context off the user stack, and start running
| it.  The program counter and status register are set by
| storing them on the supervisor stack for the "RTE" instruction.
| Remember that the order is reversed again back to normal.

ReLoad:
	moveal	A3,A0
	movel	A0,A5@(ThisTask)	| SysBase->ThisTask=newtask
	| Wait() breaks Forbid()/Disable(), so restore the old count
	movew	A0@(tc_TDNestCnt),A5@(TDNestCnt)
	movew	A0@(tc_IDNestCnt),A5@(IDNestCnt)
	movew	#TS_RUN,A0@(tc_State)	| tc_State= TS_RUN
	movel	A0@(tc_SPReg),A1	| Get the user stack pointer.

	movel	A1@+,SP@-	| Get the program counter.
	movew	A1@+,SP@-	| Get the status register.
				| The order is reversed back to normal
	movel	A1,USP		| Set user stack pointer
	moveml	A0@(tc_Regsave),D0-D7/A1-A6 | restore registers
	movel	A0@(tc_Reg_a0),A0	| Get A0
	rte			| And start running the task.

	.even
Disable:
	tstw	A5@(IDNestCnt)		| tc_IDNestCnt
	bne	DisEnd
	trap	#4			| Go to the supervisor mode
DisEnd:
	addqw	#1,A5@(IDNestCnt)	| Increase disable count
	rts
__TrapDisable:
	oriw	#0x0700,SP@		| SP|=#0x0700 Disable interrupts
	rte

	.even
Enable:
	tstw	A5@(IDNestCnt)		| test Disable count
	beq	EnEnd			| Already enabled!
	subqw	#1,A5@(IDNestCnt)
	tstw	A5@(IDNestCnt)
	bne	EnEnd
	trap	#5			| Go to the supervisor mode
EnEnd:
	rts
__TrapEnable:
	andiw	#0xf8ff,SP@		| SP&=#0xf8ff Enable interrupts
	rte

	.even
Forbid:
	FORBID
	| addqw	#1,A5@(TDNestCnt)	| Increase switch disable count
	rts

	.even
Permit:
	tstw	A5@(TDNestCnt)
	beq	EndPermit		| Already permitted
	PERMIT
	| subqw	#1,A5@(TDNestCnt)	| Decrease switch disable count
EndPermit:
	rts


	.even
Wait:					| Assembler entry point
	movel	A5@(ThisTask),A0	| ThisTask
	movel	D0,A0@(tc_SigWait)	| Parameter to tc_SigWait
	jsr	A5@(LIB_Disable)
	bra	CheckWait
WaitAgain:
	movew	#TS_WAIT,A0@(tc_State)	| tc_State=TS_WAIT
	trap	#6
	| Switcher will put us into the TaskWait-queue
	| and we return to the instruction after trap #6 when
	| someone wakes us up.
CheckWait:
	movel	A0@(tc_SigRecvd),D0	| tc_SigRecvd
	andl	A0@(tc_SigWait),D0	| mask value - signals we wait for
	| D0 = signals that caused our wakeup
	beq	WaitAgain		| If we have the signal(s), exit

	clrl	A0@(tc_SigWait)		| Waiting no more..
	eorl	D0,A0@(tc_SigRecvd)	| clear signals we waited for
	jmp	A5@(LIB_Enable)

| Signal may be called within an interrupt !! In fact it is the
| preferred way of communication between a task and a interrupt server.
	.even
Signal:					| Assembler entry point
	trap	#8
	rts
__TrapSignal:
	movew	#0x2700,SR		| Disable all interrupts.
	movel	A2,SP@-
	movel	A0,A2
	orl	D0,A2@(tc_SigRecvd)	| OR the signals
	andl	A2@(tc_SigWait),D0	| AND the new signals with tc_SigWait
	beq	EndSignal		| No wakeup
	movew	A2@(tc_State),D0
	cmpw	#TS_WAIT,D0
	bne	EndSignal		| Task is not waiting

	movel	A2,A1
	movew	#TS_READY,A1@(tc_State)	| tc_State=TS_READY
	jsr	A5@(LIB_Remove)		| Remove from TaskWait queue
	movel	A2,A1
	lea	A5@(TaskReady),A0
	jsr	A5@(LIB_Enqueue)	| Add to the TaskReady queue

| We test if the task we wakened was higher priority or not
| before we set AttnResched.
	movel	A5@(ThisTask),A0
	movel	A0,D0			| tstl A0
	beq	EndSignal
	moveb	A0@(tc_Node+ln_Pri),D0
	cmpb	A2@(tc_Node+ln_Pri),D0
	ble	EndSignal

	movew	#-1,A5@(AttnResched)	| AttnResched
EndSignal:
	movel	SP@+,A2
	rte				| resume prev. interrupt disable state

	.even
AllocSignal:
	movel	A5@(ThisTask),A0
	movel	A0@(tc_SigAlloc),D1	| Get tc_SigAlloc
	moveql	#0x1F,D0		| 32 Signals to check.  Start at 31.
loop1:
	bset	D0,D1			| Try to set the signal's flag.
	beq	found1			| If it wasn't already set, success.
	dbra	D0,loop1		| Otherwise, keep looking.
	moveq	#-1,D0			| Indicate failure.
	bra	notfound
found1:
	movel	D1,A0@(tc_SigAlloc)	| Update allocated signal flags.
notfound:
	moveq	#-1,D1
	bclr	D0,D1			| Set all but this signal bit in D1.
	andl	D1,A0@(tc_SigRecvd)	| Make sure this signal isn't set.
	andl	D1,A0@(tc_SigWait)	| Indicate we aren't waiting for it.
	rts				| Returns the bit NUMBER!! in D0

	.even
FreeSignal:
	movel	A5@(ThisTask),A0	| ThisTask
	movel	A0@(tc_SigAlloc),D1	| Get tc_SigAlloc
	bclr	D0,D1
	movel	D1,A0@(tc_SigAlloc)	| Put signals back
	rts


	.even
DateStamp:
	moveml	A5@(SystemTime),D0-D1/A1| Undividable operation - get time
	moveml	D0-D1/A1,A0@	| Store time
	rts

SetDate:
	moveml	A0@,D0-D1/A1	| Get new time
	moveml	D0-D1/A1,A5@(SystemTime) | Store new time - undividable
	rts


| Timer will not go through interrupt server, because there is unnecessary
| overhead caused by the register hassle
| This init routine SHOULD be run in supervisor mode!
	.even
inittimer:
| Prescaler 0,1/16-clock, TRR 12500, -> 200000 cycles, 100Hz
	movew	#12500,TRR1
	movew	#0x001d,TMR1
	movel	#__TimerHandler,4*(64+9) | Timer1 vector = 9
	movew	#0x0200,D0
	movew	D0,ISR		| Clear Service Flag
	orw	D0,IMR		| Enable Timer Interrupts
	rts

	.even
__TimerHandler:
	WDOG_TOGGLE		 	| Toggle the master watchdog
	movew	#0x2700,SR
	addqw	#1,A5@(IDNestCnt)	| Increase disable count

	link	A6,#0			| remember frame start
	moveml	D0-D1/A0-A3/A5,SP@-	| Save some registers (6)
	movel	AbsExecBase,A5		| Get ExecBase
	moveml	A5@(SystemTime),D0-D1/A0 | Get system time

	addql	#1,A0			| Add ticks
	cmpw	#100*60,A0		| One minute ?
	bne	LTimeOK
	movew	#0,A0			| Clear ticks
	addql	#1,D1			| Add minutes
	cmpw	#60*24,D1		| One day ?
	bne	LTimeOK
	moveql	#0,D1			| Clear minutes
	addql	#1,D0			| Add days
LTimeOK:
	moveml	D0-D1/A0,A5@(SystemTime)	| Store new time

CheckTimerRequests:
	movel	A5@(TimerRequests),A2	| lh_Head
	tstl	A2@			| lh_Head->ln_Next==0 ?
	beq	EndTimerLoop		| No requests in queue
TimerLoop:
	moveal	A2@,A3			| Remember the next node
	subql	#1,A2@(tr_Ticks)	| Decrease Ticks
	bne	NextTimer

	| OK, ticks is now zero.. return the request
	moveal	A2,A1			| Node
	jsr	A5@(LIB_Remove)		| Remove node from timerrequest-list

	movel	#SIGF_DELAY,D0		| SIGF_DELAY
	moveb	A2@(tr_Node+ln_Type),D1	| tr_Node.ln_Type
	cmpb	#NT_DELAY,D1		| NT_DELAY ?
	beq	Timersignal		| Do not return the message

	movel	A2@(tr_Task),A0		| Task *
	lea	A0@(tc_Port),A0		| tc_Port (List)
	moveal	A2,A1			| Node to add
	jsr	A5@(LIB_AddTail)	| Return the request to task's message port
	moveql	#SIGF_PORT,D0		| SIGF_PORT
Timersignal:
	moveb	#NT_REPLYMSG,A2@(tr_Node+ln_Type)
	moveal	A2@(tr_Task),A0		| A0= Task * D0= Signal to send
	jsr	A5@(LIB_Signal)

NextTimer:
	moveal	A3,A2			| Next node
	tstl	A2@			| ln_Next ? Is it a valid node
	bne	TimerLoop

EndTimerLoop:
	subqw	#1,A5@(IDNestCnt)	| Decrease disable count

	movew	A6@(4),D0		| status (no return address here!)
	andiw	#0x2000,D0		| Superuser?
	bne	EndTimer

	movel	A5@(ThisTask),A1	| ThisTask
	addql	#1,A1@(tc_Elapsed)
	subqw	#1,A1@(tc_Slice)
	bpl	SliceNotEnded		| time slice not ended ?
	movew	#-1,A5@(AttnResched)	| SysBase->AttnResched
	| clrw	A1@(tc_Slice)
SliceNotEnded:

	tstw	A5@(AttnResched)	| Check AttnReched
	beq	EndTimer
	tstw	A5@(TDNestCnt)		| TDNestCnt
	bne	EndTimer
	moveml	SP@+,D0-D1/A0-A3/A5 	| Restore registers
	unlk	A6
	moveb	#2,TER1			| Clear flag
	movew	#0x0200,ISR		| Clear Service Flag, Timer1
	bra	__Switcher
EndTimer:
	moveml	SP@+,D0-D1/A0-A3/A5 	| Restore registers
	unlk	A6
	moveb	#2,TER1			| Clear flag
	movew	#0x0200,ISR		| Clear Service Flag, Timer1
	rte


	.even
SetTaskPri:
	moveb	A0@(ln_Pri),D0	| Old pri
	moveb	D1,A0@(ln_Pri)	| set new pri (BYTE)
	extw	D0		| Extend to word
	rts

	.even
SetSignal:
	moveml	D2-D3,SP@-		| Save some registers
	jsr	A5@(LIB_Disable)
	movel	A5@(ThisTask),A0
	movel	A0@(tc_SigRecvd),D2	| Get old signals
	movel	D2,A1

	orl	D1,D2		| Clear all masked signals
	eorl	D1,D2

	movel	D1,D3
	andl	D0,D3		| New signals anded with mask will be set
	orl	D3,D2
	movel	D2,A0@(tc_SigRecvd)	| Store new signals

	jsr	A5@(LIB_Enable)
	movel	A1,D0		| Return value is original signal state
	moveml	SP@+,D2-D3 	| Restore registers
	rts

	.even
AddTask:
	movel	A5@(ThisTask),A0
	movew	A0@(tc_UserID),A1@(tc_UserID)
	movew	A0@(tc_GroupID),A1@(tc_GroupID)
	moveb	#NT_TASK,A1@(tc_Node+ln_Type)
	movew	#TS_READY,A1@(tc_State)	| tc_State=TS_READY

	lea	A1@(tc_MemList),A0
	jsr	A5@(LIB_NewList)

	movel	A1@(tc_Stack),A0	| tc_Stack
	addal	A1@(tc_StackSize),A0	| tc_Stack+tc_StackSize == end of stack
	subql	#8,A0			| Some safety margin..
	tstl	D1
	bne	NotDefault
	movel	#__Expunge,D1	| Default FinalPC is the expunge-routine
NotDefault:
	movel	A5@(ThisTask),A0@-	| Put parameter - parent Task *
	movel	D1,A0@-			| Put the FinalPC
	movew	#0,A0@-			| Status (Reversed for Switch!)
	movel	D0,A0@-			| InitialPC
	movel	A0,A1@(tc_SPReg)	| tc_SPReg

	| Initialize Child's MessagePort
	lea	A1@(tc_Port+mp_MsgList),A0	| tc_Port.mp_MsgList
	jsr	A5@(LIB_NewList)
	movel	A1,A1@(tc_Port+mp_SigTask)
	moveb	#SIGB_PORT,A1@(tc_Port+mp_SigBit)
	moveb	#PA_SIGNAL,A1@(tc_Port+mp_Flags)

	moveql	#0,D0
	movew	D0,A1@(tc_IDNestCnt) | clear interrupt and task disable counts
	movew	D0,A1@(tc_TDNestCnt)
	movel	D0,A1@(tc_SigWait)
	movel	D0,A1@(tc_SigRecvd)
	movel	D0,A1@(tc_ExceptCode)
	movel	D0,A1@(tc_Elapsed)
	movel	#0xff,A1@(tc_SigAlloc)	| SigAlloc - preallocated signals

	jsr	A5@(LIB_Disable)
	movew	A5@(Quantum),A1@(tc_Slice)| Quantum -> Elapsed counter
	lea	A5@(TaskReady),A0 | A0=List, A1=node (Task)
	jsr	A5@(LIB_Enqueue)  | Add to the TaskReady queue and return
	jsr	A5@(LIB_Enable)
	rts


	.even
.globl	__Expunge		| For dos.library ProcExpunge
__Expunge:			| Default expunge routine
	movel	AbsExecBase,A5
	movel	A5@(ThisTask),A6
	lea	A6@(tc_MemList),A0
	jsr	A5@(LIB_FreeRememberAll)
	movew	#TS_INVALID,A6@(tc_State)	| tc_State=TS_INVALID
	trap	#6				| Direct switch
| And this task is never executed again.. hopefully :-)


RemTask:
	movel	A1,D0			| Returns the task *
	FORBID
	jsr	A5@(LIB_Remove)		| Node already in A1
	PERMIT
	rts

FindTask:
	movel	A1,D0			| tstl	A1	| NULL ?
	bne	L00
	movel	A5@(ThisTask),D0
	rts
L00:
	jsr	A5@(LIB_Disable)
	lea	A5@(TaskReady),A0
	jsr	A5@(LIB_FindName)
	tstl	D0
	bne	L01
	lea	A5@(TaskWait),A0
	jsr	A5@(LIB_FindName)
L01:
	jsr	A5@(LIB_Enable)
	rts

	.even
AddIntServer:
	muluw	#lh_SIZE,D0
	lea	A5@(IntVects),A0	| &SysBase->IntVects[vector]
	addal	D0,A0			| vector List Header address
	jsr	A5@(LIB_Disable)
	jsr	A5@(LIB_AddHead)
	jsr	A5@(LIB_Enable)
	rts

	.even
RemIntServer:
	jsr	A5@(LIB_Disable)
	jsr	A5@(LIB_Remove)		| &is_Node in A1
	jsr	A5@(LIB_Enable)
	rts

	.even
Cause:
	trap	#7
	rts
__TrapCause:
	movew	#0x2700,SR		| Mask interrupts
	addqw	#1,A5@(IDNestCnt)	| Increase disable count
	moveml	D2-D7/A2-A6,SP@-	| save registers
	lea	A5@(IntVects),A4
	andl	#0xff,D0		| vector number

	muluw	#lh_SIZE,D0		| lh_SIZE*vector number
	addal	D0,A4			| SysBase->IntVects[vector#]
	movel	A4@/*(lh_Head)*/,A4
	tstl	A4@			| (is_Node+ln_Next) Is there a node ?
	beq	LCauseNo		| No nodes at all!

CauseLoop:
	movel	A4@(is_Data),D0
	movel	A4@(is_Code),A0
	movel	D0,SP@-			| is_Data to stack for C handlers
	jsr	A0@
	addql	#4,SP

	movel	A4@/*(is_Node+ln_Next)*/,A4
	tstl	A4@/*(is_Node+ln_Next)*/	| Is there a node ?
	bne	CauseLoop
LCauseNo:
	subqw	#1,A5@(IDNestCnt)	| Decrease disable count
	moveml	SP@+,D2-D7/A2-A6	| Restore registers
	rte				| Return from exception and
					| enable interrupts (rte will do)


	.even
AllocMem:
	tstl	D0
	bne	LNonZero
	rts
LNonZero:
	moveml	D2-D3,sp@-
	movel	D0,D2
	movel	D1,D3
	addql	#7,D2
	moveql	#-8,D1
	andl	D1,D2		| Aligned size

	moveql	#7,D1
	cmpl	D2,D1
	bcs	LBigEnough	| Allocated size >= sizeof(struct Memory)
	moveq	#8,D2		| D2 is aligned to 8 bytes, 8 bytes minimum
LBigEnough:
	FORBID
	lea	A5@(MemoryList),A1	| j = &SysBase->MemoryList
	moveal	A1@,A0			| i = SysBase->MemoryList
LFindLoop:
	cmpl	A0@(mm_Size),D2		| i->mm_Size < s ?
	bls	LFits
	moveal	A0,A1			| j = i;
	movel	A1@/*(mm_Next)*/,A0	| i = j->mm_Next;
	movel	A0,D0			| tstl	A0
	bne	LFindLoop
	| OK, None of the memory nodes are big enough!
	| moveql #0,D0		| Not needed because A0 was 0..
	PERMIT
	moveml	SP@+,D2-D3
	rts

LFits:
	cmpl	A0@(mm_Size),D2
	bne	LSmallerNode
	| The whole node becomes allocated. Point the mm_Next of the
	| previous node to the next node
	| In case this is the first node, SysBase->MemoryList gets changed
	movel	A0@/*(mm_Next)*/,A1@/*(mm_Next)*/
	movel	A0,D0		| And return the node pointer
	bra	LClrAlloc

LSmallerNode:
	|  ____________________       ___________________________________
	| |_n__|_s__|__________|     |_n__|_s__|_________|_n__|_s__|_____|
	|   \_________________________/_ _ _ _ _ _ _ _ _ _/
	| j                          i                   j2
	addl	D2,A1@ 		| j->mm_Next += s
	movel	A1@,A1		| j2 = j->mm_Next
	movel	A0@,A1@		| j2->mm_Next = i->mm_Next
	movel	A0@(mm_Size),D0	| j2->mm_Size = i->mm_Size -s
	subl	D2,D0
	movel	D0,A1@(mm_Size)
LClrAlloc:
	btst	#MEMB_CLEAR,D3
	beq	LClr1
	| clearing is requested
	movel	A0,A1
	lsrl	#3,D2		| Number of 2xlong's (at least 1)
	subql	#1,D2		| for dbra
	moveql	#0,D1
LClrlp:
	movel	D1,A1@+
	movel	D1,A1@+
	dbra	D2,LClrlp
	clrw	D2
	subql	#1,D2
	bcc	LClrlp
LClr1:
	PERMIT
	moveml	SP@+,D2-D3
	movel	A0,D0		| Return the start of allocated block
	rts

	.even
LNoContFree:
	rts
FreeMem:
	tstl	D0		| size == 0 ?
	beq	LNoContFree
	cmpaw	#0,A0
	beq	LNoContFree

	moveml	D2-D4/A2,SP@-
	movel	A0,D3		| Chunk address
	movel	D0,D2		| Chunk size
	addql	#7,D2
	moveq	#-8,D4
	andl	D4,D2		| D2 = s (Aligned size)
	moveq	#7,D0		| Is is aligned properly ?
	andl	D3,D0
	beq	ContFree2
	moveml	SP@+,D2-D4/A2
	rts

ContFree2:
	moveq	#7,D4
	cmpl	D2,D4
	bcs	BigEnough2
	moveq	#8,D2
BigEnough2:
	FORBID
	movel	D3,A2		| j = chunk address
	movel	A5@(MemoryList),A1	| i = SysBase->MemoryList
	cmpw	#0,A1
	beq	YesHead
	cmpl	A1,A2
	bcc	NotHead
	| The freed chunk is inserted before the first node
YesHead:
	movel	A1,A2@		| j->mm_Next = i
	movel	D2,A2@(mm_Size)	| j->mm_Size = s
	movel	A2,D0
	addl	D2,D0
	cmpl	D0,A1		| j+s == i ?
	bne	z00
	movel	A1@,A2@		| j->mm_Next = i->mm_Next
	movel	A1@(mm_Size),D4
	addl	D4,A2@(mm_Size)
z00:
	movel	A2,A5@(MemoryList)
	bra	FreeEnd

NotHead:
	movel	A1,D0		| tstl	A1
	beq	FreeEnd
	movel	A2,D1
	addl	D2,D1		| D1 = j+s
FreeLoop:
	movel	A1,D0
	addl	A1@(mm_Size),D0
	cmpl	D0,A2		| j == i + mm_Size ?
	bne	NotJustAfter	| AAAfff -> AAAAAA
	addl	D2,A1@(mm_Size)	| i->mm_Size += s
	movel	A1,D0
	addl	A1@(mm_Size),D0
	cmpl	A1@,D0		| i->mm_Next == i+i->mm_Size ?
	bne	FreeEnd		| AAAAAABBB -> AAAAAAAAA
	movel	A1@,A0		| i->mm_Next
	movel	A0@(mm_Size),D4
	addl	D4,A1@(mm_Size)	| i->mm_Size += i->mm_Next->mm_Size
	movel	A0@,A1@		| i->mm_Next = i->mm_Next->mm_Next
	bra	FreeEnd

NotJustAfter:
	tstl	A1@		| i is the last node
	bne	NotLastOne
	clrl	A2@		| j->mm_Next = 0
	movel	D2,A2@(mm_Size)	| j->mm_size = s
	movel	A2,A1@		| i->mm_Next = j
	bra	FreeEnd

NotLastOne:
	cmpl	A1@,D1		| i->mm_Next == j+s ?
	bne	NotJustBefore	| AAA fffCCC DD -> AAA CCCCCC DD
	movel	A1@,A0
	movel	A0@,A2@		| j->mm_Next = i->mm_Next->mm_Next
	addl	A0@(mm_Size),D2
	movel	D2,A2@(mm_Size)	| j->mm_Size = i->mm_Next->mm_Size + s
	movel	A2,A1@		| i->mm_Next = j
	bra	FreeEnd

NotJustBefore:
	cmpl	A1@,D1		| i->mm_Next > j+s ?
	bcc	NotBefore	| AAA fff CCC
	movel	A1@,A2@		| j->mm_Next = i->mm_Next
	movel	D2,A2@(mm_Size)	| j->mm_Size = s
	movel	A2,A1@		| i->mm_Next = j
	bra	FreeEnd

NotBefore:
	movel	A1@,A1
	movel	A1,D0		| tstl	A1
	bne	FreeLoop
FreeEnd:
	PERMIT
	moveml	SP@+,D2-D4/A2
	rts

	.even
AvailMem:
	FORBID
	movel	A5@(MemoryList),A0
	| tstl	A0
	cmpw	#0,A0			| All 32 bits are compared
	bne	ContAvail
	PERMIT
	moveql	#0,D0
	rts

ContAvail:
	btst	#MEMB_LARGEST,D0
	beq	LAvailSum
	| Search the Largest block

	moveq	#0,D0
LAvailLargest:
	cmpl	A0@(mm_Size),D0
	bgt	LAvail2
	movel	A0@(mm_Size),D0
LAvail2:
	moveal	A0@/*(mm_Next)*/,A0
	| tstl	A0
	cmpw	#0,A0			| All 32 bits are compared
	bne	LAvailLargest
	PERMIT
	rts

LAvailSum:
	moveq	#0,D0
LAvailNextSum:
	addl	A0@(mm_Size),D0
	moveal	A0@/*(mm_Next)*/,A0
	| tstl	A0
	cmpw	#0,A0			| All 32 bits are compared
	bne	LAvailNextSum
	PERMIT
	rts


AllocRemember:
	tstl	D0
	beq	LZero
	| D1 contains memory flags
	| Remember, MinNode does not have ln_Type/ln_Pri fields
	moveml	A2-A3,SP@-
	movel	A0,A3
	addl	#rm_SIZE,D0
	movel	D0,A2
	jsr	A5@(LIB_AllocMem)
	tstl	D0
	beq	LZero2

	| We can use D1 because we already allocated the memory and
	| the flags are not needed anymore..
	movel	A3,A0
	movel	A3,D1	| Tst A3
	bne	LZero3
	movel	A5@(ThisTask),A0
	lea	A0@(tc_MemList),A0	| List
LZero3:
	movel	D0,A3
	movel	A3,A1			| Node
	jsr	A5@(LIB_AddHead)	| AddHead is the fastest..
	movel	A2,A3@(rm_Size)
	movel	A3,D0
	addl	#rm_SIZE,D0
LZero2:
	moveml	SP@+,A2-A3
LZero:
	rts

FreeRemember:
	| tstl	A0
	cmpw	#0,A0			| all 32 bits are compared
	beq	LZero
	movel	A2,SP@-
	lea	A0@(-rm_SIZE),A2	| Get the node start
	movel	A2,A1
	jsr	A5@(LIB_Remove)		| Remove the node
	movel	A2@(rm_Size),D0		| The size
	movel	A2,A0			| The start address
	jsr	A5@(LIB_FreeMem)
	movel	SP@+,A2
	rts

FreeRememberAll:
	movel	A2,SP@-
	movel	A0,A2
	movel	A0,D0			| tstl A0
	bne	FreeRemLoop
	movel	A5@(ThisTask),A0	| NULL list -> tc_MemList
	lea	A0@(tc_MemList),A2	| List
FreeRemLoop:
	| Free AllocRemember():ed segments
	movel	A2,A0
	jsr	A5@(LIB_RemHead)
	tstl	D0
	beq	FreeRemEnd
	movel	D0,A0			| Start address
	movel	A0@(rm_Size),D0		| Size
	jsr	A5@(LIB_FreeMem)
	bra	FreeRemLoop
FreeRemEnd:
	movel	SP@+,A2
	rts

AddProgram:
	moveb	#NT_RESIDENT,A1@(pm_Node+ln_Type)
	FORBID
	lea	A5@(ProgramList),A0	| List to A0, Node in A1
	jsr	A5@(LIB_AddTail)
	PERMIT
	rts

RemProgram:
	FORBID
	lea	A5@(ProgramList),A0	| List to A0, Name in A1
	jsr	A5@(LIB_FindName)
	tstl	D0
	beq	LRemLeave
	movel	D0,A1
	jsr	A5@(LIB_Remove)		| Does not mess D0
LRemLeave:
	PERMIT
	rts

FindProgram:
	FORBID
	lea	A5@(ProgramList),A0	| List to A0, Name in A1
	jsr	A5@(LIB_FindName)
	PERMIT
	rts

CopyMem:	| A1 -> A0, D0 bytes, uses D1, A0&A1 word-aligned!
	movel	D0,D1
	lsrl	#2,D1			| size/4 == copy LONGs
	subql	#1,D1
Lcpmemloop:
	movel	A1@+,A0@+
	dbra	D1,Lcpmemloop
	btst	#1,D0
	beq	Lcpmem0
	movew	A1@+,A0@+
Lcpmem0:
	btst	#0,D0
	beq	Lcpmem1
	moveb	A1@+,A0@+
Lcpmem1:
	rts


AddLibrary:
	moveb	#NT_LIBRARY,A1@(lib_Node+ln_Type)
	FORBID
	lea	A5@(LibList),A0
	jsr	A5@(LIB_AddTail)
	PERMIT
	rts

RemLibrary:
	FORBID
	tstw	A1@(lib_OpenCnt)
	bne	LNotRemLib
	movel	A1,D0
	jsr	A5@(LIB_Remove)	| Does not mess D0
	PERMIT
	rts
LNotRemLib:
	PERMIT
	moveql	#0,D0
	rts

InitLibFuncs:	| Lib * = A0, funcs * = A1, number of funcs = D0
	tstl	D0
	bne	InitLibFuncsLoop
	rts
InitLibFuncsLoop:
	movel	A1@+,A0@-	| Move the jump address
	movew	#0x4ef9,A0@-	| Move the jump instruction
	subql	#1,D0
	bne	InitLibFuncsLoop
	rts
	
OpenLibrary:
	FORBID
	lea	A5@(LibList),A0
	jsr	A5@(LIB_FindName)
	tstl	D0
	bne	L30
	PERMIT
	rts
L30:
	movel	D0,A5
	jsr	A5@(LIB_Open)
	| D0 is the library base for successful opens
	movel	AbsExecBase,A5
	PERMIT
	rts

CloseLibrary:
	FORBID
	movel	A1,A5
	jsr	A5@(LIB_Close)
	movel	AbsExecBase,A5
	PERMIT
	rts

SetFunction:
	| A1 = library *, D0 = vector offset (negative), D1 = new value
	addl	D0,A1
	FORBID
	movel	A1@(2),D0
	movel	D1,A1@(2)
	PERMIT
	rts

AddResource:
	moveb	#NT_RESOURCE,A1@(lib_Node+ln_Type)
	FORBID
	lea	A5@(ResourceList),A0
	jsr	A5@(LIB_AddTail)
	PERMIT
	rts

RemResource:
AllocResource:
	FORBID
	lea	A5@(ResourceList),A0
	jsr	A5@(LIB_FindName)
	PERMIT
	rts

FreeResource:
	FORBID
	lea	A5@(ResourceList),A0
	jsr	A5@(LIB_AddTail)
	PERMIT
	rts

strNoMemory:
	.asciz	"No Memory!\n\r"
strTimerOK:
	.asciz	"Timer OK\n\r"
strMemoryListOK:
	.asciz	"MemList OK\n\r"
strStackOK:
	.asciz	"Stack OK   \n\r"
ExecName:
	.asciz	"exec.library"
strDeadEnd:
	.asciz	"*Deadend Alert!*\n\r"
strProgramError0:
	.asciz	"** Program Error **\n\rCause: "
strProgramError1:
	.asciz	"\n\rTask: "
newline:
	.asciz	"\n\r"
strPC:
	.asciz	"Program Counter: "
strSR:
	.asciz	"Status Register: "
strIR:
	.asciz	"Instruction    : "
strAccess:
	.asciz	"Access Address : "
strFC:
	.asciz	"Function Code  : "
strExecBase:
	.asciz	"ExecBase: "
	
strRes:	.asciz	"(Unassigned, Reserved)"
str00:	.asciz	"Reset (SP)"
str01:	.asciz	"Reset"
str02:	.asciz	"Bus Error"
str03:	.asciz	"Address Error"
str04:	.asciz	"Illegal Instruction"
str05:	.asciz	"Zero Divide"
str06:	.asciz	"CHK Instruction"
str07:	.asciz	"TRAPV Instruction"
str08:	.asciz	"Privilege Violation"
str09:	.asciz	"Trace"
str10:	.asciz	"Line-A Emulator"
str11:	.asciz	"Line-F Emulator"
str15:	.asciz	"Uninitialized Interrupt Vector"
str24:	.asciz	"Spurious Interrupt"
str25:	.asciz	"Level 1 Autovector"
str26:	.asciz	"Level 2 Autovector"
str27:	.asciz	"Level 3 Autovector"
str28:	.asciz	"Level 4 Autovector"
str29:	.asciz	"Level 5 Autovector"
str30:	.asciz	"Level 6 Autovector"
str31:	.asciz	"Level 7 Autovector"
str32:	.asciz	"Trap #0"
str33:	.asciz	"Trap #1"
str34:	.asciz	"Trap #2"
str35:	.asciz	"Trap #3"
str36:	.asciz	"Trap #4"
str37:	.asciz	"Trap #5"
str38:	.asciz	"Trap #6"
str39:	.asciz	"Trap #7"

	.even
strPtrInt:
	.long	str00,str01,str02,str03,str04,str05,str06,str07
	.long	str08,str09,str10,str11,strRes,strRes,strRes,str15
	.long	strRes,strRes,strRes,strRes,strRes,strRes,strRes,strRes
	.long	str24,str25,str26,str27,str28,str29,str30,str31
	.long	str32,str33,str34,str35,str36,str37,str38,str39

	.even
ExecVersion:
	.word	3

