#ifndef DOS_PROTOS_H
#define DOS_PROTOS_H

#include <exec/types.h>
#include <dos/dos.h>
#include <stdarg.h>

struct MsgPort *DeviceProc(const char *handler);
LONG DoPkt4(struct MsgPort *port, LONG type, ULONG arg1,
		   ULONG arg2, ULONG arg3, ULONG arg4);

LONG DoPkt4Safe(const char *handler, LONG type, ULONG arg1,
		       ULONG arg2, ULONG arg3, ULONG arg4);

APTR  Open(const char *filename, ULONG mode);
LONG  Close(APTR handle);
LONG  Read(APTR handle, char *buffer, LONG size);
LONG  Write(APTR handle, const char *buffer, LONG size);
LONG  SetMode(APTR handle, LONG mode);
APTR  Input(void);
APTR  Output(void);
APTR  Error(void);
LONG  Seek(APTR handle, LONG offset, LONG direction);
LONG  DeleteFile(const char *file);
LONG  Rename(const char *oldName, const char *newName);
APTR  Lock(const char *name, LONG mode);
LONG  UnLock(APTR lock);
APTR  DupLock(APTR oldLock);
LONG  SameLock(APTR lockA, APTR lockB);
LONG  Examine(APTR lock, struct FileInfoBlock *fib);
LONG  ExNext(APTR lock, struct FileInfoBlock *fib);
LONG  ExamineEnd(APTR lock, struct FileInfoBlock *fib);
APTR  CreateDir(const char *name);
APTR  CurrentDir(APTR newCurrentDir);
APTR  ParentDir(APTR lock);
APTR  SetComment(const char *name, const char *comment);
APTR  SetProtection(const char *name, ULONG protection);
ULONG SetOwner(const char *name, ULONG ownerinfo);
LONG  Format(const char *disk, ULONG flags);
LONG  DiskInfo(APTR lock, struct InfoData *info);
LONG  IoErr(void);
void  SetIoErr(LONG);

struct SegList *ROMLoadSeg(const char *name);
struct SegList *LoadSeg(const char *name);
LONG   UnLoadSeg(struct SegList *segList);
LONG RunCommand(struct SegList *segList,
		ULONG stacksize,
		const char *args,
		ULONG arglen);
struct Process *CreateProc(const char *name,
			   LONG pri,
			   struct SegList *segList,
			   ULONG stackSize,
			   const char *argString);

long WaitForChar(APTR handle, ULONG time);
LONG DeleteVar(const char *name, ULONG flags);
struct LocalVar *FindVar(const char *name, ULONG flags);
LONG GetVar(const char *name, char *buffer, LONG bufsize, ULONG flags);
LONG SetVar(const char *name, const char *buffer, LONG size, ULONG flags);

LONG  PrintFault( LONG code, const char *header );

void Delay( ULONG ticks );

/* New */
LONG SetFileDate( const char *name, const struct DateStamp *date );
LONG IsFileSystem( const char *handlername );
LONG IsInteractive( APTR handle );
LONG Relabel( const char *handler, const char *newname );

struct Process *CreateNewProcTags( ULONG tag1type, ... );
LONG SystemTags( const char *command, ULONG tag1type, ... );

struct DevProc *GetDeviceProc(const char *handler, struct DevProc *old);
void FreeDeviceProc(struct DevProc *dev);

LONG AssignLock(const char *assign, APTR lock);
LONG AssignLate(const char *assign, const char *path);
LONG AssignAdd(const char *assign, APTR lock);

LONG PrintFmt(va_list ap, void *fp, int (*cf)(int, void *), const char *fmt);
LONG FPrintF(APTR handle, const char *fmt, ...);
LONG PrintF(const char *fmt, ...);

struct ArgList *ParseArgs(const char *);
void FreeArgs(struct ArgList *);

#endif
