#ifndef TIME_H
#define TIME_H

#include "stddef.h"

typedef unsigned long clock_t;
typedef long time_t;

struct tm {
    int tm_sec;   /* seconds after the minute (0 .. 59)  */
    int tm_min;   /* minutes after the hour   (0 .. 59)  */
    int tm_hour;  /* hours since midnight     (0 .. 23)  */
    int tm_mday;  /* day of month             (0 .. 31)  */
    int tm_mon;   /* months since January     (0 .. 11)  */
    int tm_year;  /* years since 1900                    */
    int tm_wday;  /* days since Sunday        (0 .. 6)   */
    int tm_yday;  /* days since Jan 1         (0 .. 365) */
    int tm_isdst; /* Daylight Savings Time flag */
};

#define CLK_TCK 100

clock_t clock(void);
time_t time(time_t *tp);
/* NOTE! difftime is NOT pure ANSI. In ANSI, the return type of the */
/*       functions is double, but since we don't have doubles here, */
/*       this implementation returns only a long unsigned int. This */
/*       will cause the fact that very big time differences (> 78  */
/*       years can't be properly calculated!                        */
long difftime(time_t time2, time_t time1);
time_t mktime(struct tm *tp);
char *asctime(const struct tm *tp);
#define ctime(tp) (asctime(localtime(tp)))
struct tm *gmtime(const time_t *tp);
struct tm *localtime(const time_t *tp);
size_t strftime(char *s, size_t smax, const char *fmt, const struct tm *tp);

#ifndef __STRICT_ANSI
       /* convert time to OS DateStamp, returns NULL on error */
struct DateStamp *tmtods(const struct tm *tp);
       /* returns current time zone in minutes */
       /* defaults to 0 if not available       */
int tzminutes(void);
#endif

/* Private structures, variables and functions start here. DO NOT USE! */

extern unsigned long __starttime;
extern struct tm __systm;
extern const char __mnames[12][4];
extern const char __monthnames[12][10];
extern const char __dnames[7][4];
extern const char __weekdaynames[7][10];
extern const unsigned short __ydays[2][12];

#endif
