#include "stdio.h"
#include "string.h"
#include "ctype.h"
#include "stdarg.h"
#include <dos/dos.h>

#define F_ALIGN_LEFT         1
#define F_ALWAYS_SIGN        2
#define F_NEGATIVE           4
#define F_SPACE_IF_NOT_SIGN  8
#define F_FILL_ZERO         16
#define F_ALTERNATE         32


#define GETNEXT(a, b) {if(!((a) = (*((b)++)))) {va_end(ap); return -1;}}

static const char hex[16]="0123456789abcdef";

static FILE *acfp;
static int acwritten;
static int (*putcfunc)(int, FILE *);

static void addchar(char c)
{
    putcfunc(c, acfp);
}

static int add(char *src, int n, int w, ULONG flags)
{
    int i;
    char leftfillchar = (flags &  F_FILL_ZERO) ? '0' : ' ';

    w -= n;
    if (flags & (F_NEGATIVE | F_ALWAYS_SIGN | F_SPACE_IF_NOT_SIGN))
	w--;

    acwritten = 0;
    if (!(flags & F_ALIGN_LEFT) && leftfillchar == ' ')
    {
	if (w > 0)
	{
	    acwritten += w;
	    while (w-- > 0)
		addchar(leftfillchar);
	}
    }

    if (flags & (F_NEGATIVE | F_ALWAYS_SIGN | F_SPACE_IF_NOT_SIGN))
    {
	acwritten++;
	if (flags & F_NEGATIVE)
	{
	    addchar('-');
	}
	else
	{
	    if (flags & F_ALWAYS_SIGN)
		addchar('+');
	    else
		addchar(' ');
	}
    }

    if (!(flags & F_ALIGN_LEFT) && leftfillchar == '0')
	if (w > 0)
	{
	    acwritten += w;
	    while (w-- > 0)
		addchar(leftfillchar);
	}

    acwritten += n;
    for (i=0; i<n; i++)
	addchar(*src++);

    if (flags & F_ALIGN_LEFT)
	if (w > 0)
	{
	    acwritten += w;
	    while (w-- > 0)
		addchar(' ');
	}

    return acwritten;
}



int __doscanf(va_list ap, FILE *fp, int (*cf)(int, FILE *), const char *fmt)
{
    long value;
    ULONG uvalue;
    char *svalue;
    char curr;
    int written = 0;

    putcfunc = cf;
    acfp = fp;

    curr = *fmt++;
    while (curr)
    {
	if (curr == '%')
	{
	    BOOL currucase;
	    int width = 0;
	    char tmpstr[20];
	    int i;
	    ULONG flags = 0;
	    BOOL phase1 = TRUE;
	    BOOL dont_get = FALSE;

	    while (phase1) {
		if (!dont_get) GETNEXT(curr, fmt);
		dont_get = FALSE;

		switch(curr)
		{
		case '-':
		    flags |= F_ALIGN_LEFT;
		    break;
		case '+':
		    flags |= F_ALWAYS_SIGN;
		    break;
		case ' ':
		    flags |= F_SPACE_IF_NOT_SIGN;
		    break;
		case '0':
		    flags |= F_FILL_ZERO;
		    break;
		case '#':
		    flags |= F_ALTERNATE;
		    break;
		default:
		    if (isdigit(curr))
		    {
			while (isdigit(curr))
			{
			    width = width*10 + (curr - '0');
			    GETNEXT(curr, fmt);
			}
			dont_get = TRUE;
		    }
		    else
		    {
			phase1 = FALSE;
		    }
		    break;
		}		
	    }

	    if (tolower(curr) == 'l' || curr == 'h')
		GETNEXT(curr, fmt);

	    currucase = isupper(curr);
	    switch(tolower(curr)) {
	    case 'i':
	    case 'd':
		value = va_arg(ap, long);
		i = 0;
		if (value == 0x80000000)
		{
		    flags |= F_NEGATIVE;
		    memcpy(&tmpstr[19-(10-1)-i], "2147483648", 10);
		    i += 10;
		}
		else
		{
		    if (value < 0)
		    {
			flags |= F_NEGATIVE;
			value = -value;
		    }
		    while (value)
		    {
			tmpstr[19-i++] = value % 10 + '0';
			value /= 10;
		    }
		}
		if (!i) tmpstr[19-i++] = '0';
		written += add(tmpstr+20-i, i, width, flags);
		break;
	    case 'u':
		uvalue = va_arg(ap, long);
		i = 0;
		while (uvalue)
		{
		    tmpstr[19-i++] = uvalue % 10 + '0';
		    uvalue /= 10;
		}
		if (!i) tmpstr[19-i++] = '0';
		written += add(tmpstr+20-i, i, width, flags);
		break;
	    case 'o':
		uvalue = va_arg(ap, long);
		i = 0;

		while (uvalue)
		{
		    tmpstr[19-i++] = (uvalue & 7) + '0';
		    uvalue >>= 3;
		}
		if (!i || flags & F_ALTERNATE)
		    tmpstr[19-i++] = '0';

		written += add(tmpstr+20-i, i, width, flags);
		break;
	    case 'p':
	    case 'x':
		uvalue = va_arg(ap, unsigned long);
		i = 0;
		while (uvalue)
		{
		    tmpstr[19-i++] = (currucase) ?
			toupper(hex[uvalue & 15]) :
			hex[uvalue & 15];
		    uvalue >>= 4;
		}
		if (i && flags & F_ALTERNATE)
		{
		    tmpstr[19-i++] = curr;
		    tmpstr[19-i++] = '0';
		}
		if (!i) tmpstr[19-i++] = '0';

		written += add(tmpstr+20-i, i, width, flags);
		break;
	    case 'c':
		tmpstr[0] = va_arg(ap, char);
		add(tmpstr, 1, 1, 0);
		break;
	    case 's':
		svalue = va_arg(ap, char *);
		flags &= ~(F_FILL_ZERO | F_ALWAYS_SIGN | F_SPACE_IF_NOT_SIGN);
		written += add(svalue, strlen(svalue), width, flags);
		break;
	    case 'f':
	    case 'e':
	    case 'g':
		written += add("<float>", 7, 7, 0);
		break;
	    case '%':
		tmpstr[0] = '%';
		add(tmpstr, 1, 1, 0);
		break;
	    default:
		return -1;   /* Unrecognized code, aborting */
		break;
	    }
	}
	else
	{
	    written++;
	    putcfunc(curr, fp);
	}

	curr = *fmt++;
    }

    va_end(ap);

    return written;
}


