/* Actually, this is a merge sort to avoid excessive stack usage */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

void qsort(void *base, size_t n, size_t size,
	   int (*cmp)(const void *, const void *))
{
    long i, j;
    void *tmp = malloc(size);

    if (tmp == NULL)
	return;

    for (i = 2; (i >> 1) < n; i *= 2)         /*  merge sets of 2,4,8.. */
    {
	for (j = 0; j < n; j += i)
	{
	    long i1 = i >> 1;
	    long i2 = i1;
	    char *p1;
	    char *p2;

	    if (size == 4)
	    {
		p1 = (char *)base + j * 4;
		p2 = p1 + i1 * 4;
	    }
	    else
	    {
		p1 = (char *)base + j * size;
		p2 = p1 + i1 * size;
	    }

	    if (j + i1 + i2 > n)
		i2 = n - i1 - j;

	    /*
	     *	Merge p2/i2 into p1/i1 until p2/i2 is exhausted.  Note that
	     *	i2 might start out negative (in which case i1 is invalid,
	     *	but the segment is already sorted anyway)
	     */

	    while (i2 > 0 && i1 > 0)
	    {
		/*
		 *  compare current p1/i1 with next p2/i2.  Skip p1 while
		 *  p1 < p2.  Otherwise, merge the p2/i2 element into
		 *  p1/i1.
		 */

		if ((*cmp)(p1, p2) <= 0)
		{
		    i1--;
		    p1 += size;
		}
		else
		{
		    if (size == 4)
		    {
			*(long *)tmp = *(long *)p2;
			memmove(p1 + 4, p1, i1 * 4);
			*(long *)p1 = *(long *)tmp;
		    }
		    else
		    {
			memmove(tmp, p2, size);
			memmove(p1 + size, p1, i1 * size);
			memmove(p1, tmp, size);
		    }
		    i2--;
		    p2 += size;
		    p1 += size;	/*  shifted p1, i1 stays    */
		}
	    }
	}
    }
    free(tmp);
}

