/*#include "ctype.h"
#include "stdarg.h"*/

#include "stdio.h"
#include "time.h"
#include "string.h"
#include <dos/dos.h>


static int acwritten, acmax;
static char *acbuf, *aclast;

static void addinit(char *buf, int max)
{
    acmax = max;
    acbuf = buf;
    aclast = acbuf + acmax - 1;
    acwritten = 0;
}

static void addchar(char c)
{
    if (acwritten++ < acmax)
	*acbuf++ = c;
}

static void add(const char *src)
{
    while (*src)
	addchar(*src++);
}

static int addend(void)
{
    if (acwritten >= acmax)
    {
	*aclast = '\0';
	return 0;
    }
    else
    {
	*acbuf = '\0';
	return acwritten+1;
    }
}


size_t strftime(char *d, size_t smax, const char *fmt, const struct tm *tp)
{
    char curr;
    static char tmpstr[20];

    addinit(d, smax);

    curr = *fmt++;
    while (curr)
    {
	if (curr == '%')
	{
	    curr = *fmt++;
	    if (!curr) {
		addchar('#');
		goto endit;
	    }

	    switch(curr) {
	    case 'a':
		add(__dnames[tp->tm_wday]);
		break;
	    case 'A':
		add(__weekdaynames[tp->tm_wday]);
		break;
	    case 'b':
		add(__mnames[tp->tm_mon]);
		break;
	    case 'B':
		add(__monthnames[tp->tm_mon]);
		break;
	    case 'c':
		sprintf(tmpstr, "%4d-%02d-%02d %2d:%02d:%02d",
			tp->tm_year+1900, tp->tm_mon+1, tp->tm_mday+1,
			tp->tm_hour, tp->tm_min, tp->tm_sec);
		add(tmpstr);
		break;
	    case 'd':
		sprintf(tmpstr, "%02d", tp->tm_mday+1);
		add(tmpstr);
		break;
	    case 'H':
		sprintf(tmpstr, "%02d", tp->tm_hour);
		add(tmpstr);
		break;
	    case 'I':
		sprintf(tmpstr, "%02d", ((tp->tm_hour + 11) % 12) + 1);
		add(tmpstr);
		break;
	    case 'j':
		sprintf(tmpstr, "%03d", tp->tm_yday+1);
		add(tmpstr);
		break;
	    case 'm':
		sprintf(tmpstr, "%02d", tp->tm_mon+1);
		add(tmpstr);
		break;
	    case 'M':
		sprintf(tmpstr, "%02d", tp->tm_min);
		add(tmpstr);
		break;
	    case 'p':
		add((tp->tm_hour < 12) ? "AM" : "PM");
		break;
	    case 'S':
		sprintf(tmpstr, "%02d", tp->tm_sec);
		add(tmpstr);
		break;
	    case 'U':
		add("**");   /*** WEEK # (Sun = 1st day) NOT SUPPORTED YET! **/
		break;
	    case 'W':
		add("**");   /*** WEEK # (Mon = 1st day) NOT SUPPORTED YET! **/
		break;
	    case 'x':
		sprintf(tmpstr, "%4d-%02d-%02d",
			tp->tm_year+1900, tp->tm_mon+1, tp->tm_mday+1);
		add(tmpstr);
		break;
	    case 'X':
		sprintf(tmpstr, "%2d:%02d:%02d",
			tp->tm_hour, tp->tm_min, tp->tm_sec);
		add(tmpstr);
		break;
	    case 'y':
		sprintf(tmpstr, "%02d", tp->tm_year % 100);
		add(tmpstr);
		break;
	    case 'Y':
		sprintf(tmpstr, "%4d", tp->tm_year + 1900);
		add(tmpstr);
		break;
	    case 'Z':
		add("EET");          /***** TIME ZONE NOT SUPPORTED YET! *****/
		break;
	    case '#':
		addchar('#');
		break;
	    default:
		addchar('#');       /* This should actually be an error */
		addchar(curr);
	    }
	}
	else
	{
	    addchar(curr);
	}

	curr = *fmt++;
    }

 endit:
    return addend();
}
