#ifndef FORTUNE_FILE
#define FORTUNE_FILE "ROM:fortune/Fortunes"
#endif

/*
 * Fortune file search order
 *
 * Amiga:						Other:
 * ~~~~~~						~~~~~~
 * Implicitly named				Implicitly named
 * Fortunes:Fortunes
 * $HOME/.Fortunes (local var)	$HOME/.Fortunes
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* #define CONVERSIONS */

#define FILENAMESIZE	256
#define LINESIZE		100
#define MAXSIZE			2000
#define MAXFORTUNES		4000

#define OPT_USAGE	(1<<0)
#define OPT_INDEX	(1<<1)
#define OPT_GREP	(1<<2)
#ifdef CONVERSIONS
#define OPT_CONV	(1<<3)
#define OPT_CONVOLD	(1<<4)
#endif
#define OPT_SHORT	(1<<5)
#define OPT_NAMED	(1<<6)

static int exists(char *file);
static int exists(char *file)
{
    FILE *fp;
    
    if((fp = fopen(file, "rb")))
    {
	fclose(fp);
	return -1;
    }
    return 0;
}


int main(int argc, char *argv[])
{
    FILE *fp;
    char buffer[LINESIZE+2], fortune[MAXSIZE], greppat[LINESIZE+2];

    char FortuneFile[FILENAMESIZE] = "Fortunes:Fortunes";
    char NameBuffer[FILENAMESIZE+8];

#ifdef CONVERSIONS
    char ConvFile[FILENAMESIZE], OldConvFile[FILENAMESIZE];
#endif

    int i;
    long options = 0, *positions, n, *fourpos, fourn, size, z=-1;
    long fortunestart, fortunelines;
    
    fortune[0] = '\0';
    for(n=1;n<argc;n++)
    {
	if(argv[n][0]=='-')
	{
	    i=1;
	    while(argv[n][i])
	    {
		switch(argv[n][i])
		{
		case 'i':
		    options |= OPT_INDEX;
		    break;
		case 'g':
		    options |= OPT_GREP;
		    strncpy(greppat,&argv[n][i+1],LINESIZE);
		    i = strlen(argv[n])-1;
		    break;
#ifdef CONVERSIONS
		case 'c':
		    options |= OPT_CONV;
		    strncpy(ConvFile,&argv[n][i+1],LINESIZE);
		    i = strlen(argv[n])-1;
		    break;
		case 'o':
		    options |= OPT_CONVOLD;
		    strncpy(OldConvFile,&argv[n][i+1],LINESIZE);
		    i = strlen(argv[n])-1;
		    break;
#endif
		case 's':
		    options |= OPT_SHORT;
		    break;
		case 'n':
		    z = atol(&argv[n][i+1]);
		    i = strlen(argv[n])-1;
		    break;
		default:
		    options |= OPT_USAGE;
		}
		i++;
	    }
	}
	else
	{
	    if(!(options & OPT_NAMED))
	    {
		strncpy(FortuneFile, argv[n], FILENAMESIZE);
		options |= OPT_NAMED;
	    }
	    else
	    {
		options |= OPT_USAGE;
	    }
	}
    }
    if(options & OPT_USAGE)
    {
#ifdef CONVERSIONS
	printf(
	       "Usage: %s [-i] [-s] [-n<number>] [-c<inputfile>] [-o<inputfile>] [-g<grep pattern>] [<FortuneFile>]\n"
	       "       -i  Generate an index file\n"
	       "       -n  Show a specified fortune\n"
	       "       -g  Show all quotes including the specified string\n"
	       "       -s  display short fortunes (max 4 lines)\n"
	       "       -c  Convert 'old type' fortune file\n"
	       "       -o  Convert to 'old type'\n", argv[0]);
#else
	printf(
	       "Usage: %s [-i] [-s] [-n<number>] [-g<grep pattern>] [<FortuneFile>]\n"
	       "       -i  Generate an index file\n"
	       "       -n  Show a specified fortune\n"
	       "       -g  Show all quotes including the specified string\n"
	       "       -s  display short fortunes (max 4 lines)\n", argv[0]);
#endif
	exit(5);
    }

    /* if no fortune file is implicitly named, decide which file to use.. */
    if(!(options & OPT_NAMED))
    {
	char *env;
	
	if((env = getenv("HOME")))
	{
	    char z = env[strlen(env)-1];
	    
	    strcpy(FortuneFile, env);
	    free(env);
	    if(z != '/' && z != ':')
	    {
		strcat(FortuneFile, "/");
	    }
	    strcat(FortuneFile,".Fortunes");
	    if(exists(FortuneFile))
	    {
		options |= OPT_NAMED;
	    }
	}
	if(!(options & OPT_NAMED))
	{
	    strcpy(FortuneFile, FORTUNE_FILE);
	    if(exists(FortuneFile))
	    {
		options |= OPT_NAMED;
	    }
	}
	
	if(!(options & OPT_NAMED))
	{
	    fprintf(stderr, "Could not find the fortune file!\n");
	    exit(10);
	}
    }

#ifdef CONVERSIONS
    if(options & OPT_CONVOLD)
    {
	if((fp=fopen(OldConvFile,"rb")))
	{
	    n=0;
	    while(fgets(buffer,LINESIZE,fp))
	    {
		if(!strncmp(buffer,"%%",2))
		{
		    n++;
		}
		else
		{
		    printf("%5d%%%s",n,buffer);
		}
	    }
	    fclose(fp);
	}
	else
	{
	    printf("Could not open fortune file: \"%s\".\n",OldConvFile);
	    exit(10);
	}
    }
    if(options & OPT_CONV)
    {
	if((fp=fopen(ConvFile,"rb")))
	{
	    fortune[0] = '\0';
	    while(fgets(buffer,LINESIZE,fp))
	    {
		if(!strncmp(buffer,"%%",2))
		{
		    printf("The file \"%s\" is already a new format fortune file!\n", greppat);
		    break;
		}
		i=0;
		while(buffer[i]!='%' && buffer[i])	i++;
		if(!buffer[i])
		    continue;
		
		if(strncmp(fortune,buffer,i))
		{
		    if(fortune[0]!='\0')
		    {
			printf("%%%%\n%s",buffer+i+1);
		    }
		    strncpy(fortune,buffer,i);
		}
		else
		{
		    printf("%s",buffer+i+1);
		}
	    }
	    printf("%%%%\n");
	    fclose(fp);
	}
	else
	{
	    printf("Could not open fortune file: \"%s\".\n",ConvFile);
	    exit(10);
	}
    }
#endif

    if(options & OPT_GREP)
    {
	if(fp=fopen(FortuneFile,"rb"))
	{
	    while(fgets(buffer,LINESIZE,fp))
	    {
		if(!strncmp(buffer,"%%",2))
		{
		    if(strstr(fortune,greppat))
			printf("%s%%%%\n",fortune);
		    fortune[0] = '\0';
		}
		else
		{
		    if(strlen(fortune)+strlen(buffer)<MAXSIZE)
			strcat(fortune,buffer);
		    else
			fprintf(stderr, "Error: too long fortune\n");
		}
	    }
	    fclose(fp);
	}
	else
	{
	    printf("Could not open fortune file :\"%s\".\n",FortuneFile);
	    exit(10);
	}
    }
    
    if(options & OPT_INDEX)
    {
	fourpos = (long *)calloc(2*MAXFORTUNES, sizeof(long));
	if(positions = (long *)calloc(2*MAXFORTUNES, sizeof(long)))
	{
	    if(fp=fopen(FortuneFile,"rb"))
	    {
		n = 0;
		fourn = 0;
		fortunelines = 0;
		fortunestart = ftell(fp);
		while(n<2*MAXFORTUNES)
		{
		    if(!fgets(buffer,LINESIZE,fp))
			break;
		    if(!strncmp(buffer,"%%",2))
		    {
			positions[n++] = fortunestart;
			positions[n++] = strlen(fortune);
			
			if(fourpos && fortunelines<5)
			{
			    fourpos[fourn++] = fortunestart;
			    fourpos[fourn++] = strlen(fortune);
			}
			
			fortunestart = ftell(fp);
			fortune[0] = '\0';
			fortunelines = 0;
		    }
		    else
		    {
			if(strlen(fortune)+strlen(buffer)<MAXSIZE)
			    strcat(fortune,buffer);
			else
			    fprintf(stderr, "Error: too long fortune %d\n",n/2);
			fortunelines++;
		    }
		}
		n >>= 1;
		fourn >>= 1;
		fclose(fp);
		
		sprintf(NameBuffer, "%s.idx", FortuneFile);
		if(fp=fopen(NameBuffer, "wb"))
		{
		    fwrite((char *)&n, 1, sizeof(long), fp);
		    fwrite((char *)&positions[0], 2*n, sizeof(long), fp);
		    fclose(fp);
		    printf("Index file with %d fortunes created.\n", n);
		}
		else
		{
		    printf("Index file generation failed (%d fortunes).\n", n);
		}
		sprintf(NameBuffer, "%s.idx4", FortuneFile);
		if(fourpos && (fp=fopen(NameBuffer, "wb")))
		{
		    fwrite((char *)&fourn, 1, sizeof(long), fp);
		    fwrite((char *)&fourpos[0], 2*fourn, sizeof(long), fp);
		    fclose(fp);
		    printf("Short");
		    printf("Index file with %d fortunes created.\n", fourn);
		}
		else
		{
		    printf("Short");
		    printf("Index file generation failed (%d fortunes).\n", fourn);
		}
	    }
	    else
	    {
		printf("Could not open fortune file :\"%s\".\n",FortuneFile);
		free(positions);
		if(fourpos)
		{
		    free(fourpos);
		}
		exit(10);
	    }
	    free(positions);
	}
	else
	{
	    printf("Memory allocation failed\n");
	}
	if(fourpos)
	{
	    free(fourpos);
	}
    }

    if(options & OPT_SHORT)
    {
	sprintf(NameBuffer, "%s.idx4", FortuneFile);
	if(fp=fopen(NameBuffer, "rb"))
	{
	    fread((char *)&n, 1, sizeof(long), fp);
	    if(z < 0 || z > n)
	    {
		srand(time(NULL));
		z = (rand()/2) % n;
	    }
	    fseek(fp, 2*z*sizeof(long), SEEK_CUR);
	    fread((char *)&n, 1, sizeof(long), fp);
	    fread((char *)&size, 1, sizeof(long), fp);
	    fclose(fp);
	    if(fp=fopen(FortuneFile,"rb"))
	    {
		fseek(fp, n, SEEK_SET);
		fread(fortune, size, sizeof(char), fp);
		fortune[size] = '\0';
		printf("%s",fortune);
		fclose(fp);
	    }
	}
	else
	{
	    printf("Could not open index file :\"%s\".\n", NameBuffer);
	}
    }
    if(!(options & ~OPT_NAMED))
    {
	sprintf(NameBuffer, "%s.idx", FortuneFile);
	if(fp=fopen(NameBuffer, "rb"))
	{
	    fread((char *)&n, 1, sizeof(long), fp);
	    if(z < 0 || z > n)
	    {
		srand(time(NULL));
		z = (rand()/2) % n;
	    }
	    fseek(fp, 2*z*sizeof(long), SEEK_CUR);
	    fread((char *)&n, 1, sizeof(long), fp);
	    fread((char *)&size, 1, sizeof(long), fp);
	    fclose(fp);
	    if(fp=fopen(FortuneFile,"rb"))
	    {
		fseek(fp, n, SEEK_SET);
		fread(fortune, size, sizeof(char), fp);
		fortune[size] = '\0';
		printf("\n%s\n",fortune);
		fclose(fp);
	    }
	}
	else
	{
	    printf("Could not open index file :\"%s\".\n", NameBuffer);
	}
    }
    return 0;
}
