#include <dos/dos_protos.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "crctab.c"

void __chkabort(void) {}

#define NUL 0x00
#define SOH 0x01
#define STX 0x02
#define EOT 0x04
#define ACK 0x06
#define NAK 0x15
#define CAN 0x18
#define MAXRETRIES	10
#define MAXBLOCKSIZE	1024

int getkey(APTR input, unsigned long time)
{
    unsigned char q;

    if(WaitForChar(input, time))
    {
	if(Read(input, (char *)&q, 1))
	{
	    return q;
	}
    }
    return -1;
}

static int receiveblock(APTR input, APTR output, unsigned char *block, int expblocknum, int *size, int crcmode)
{
    unsigned char keys[2];
    unsigned int sum, crc;
    int blokki, i=0, j;

    while(1)
    {
	/* wait for SOH/STX/EOT (10 secs) */

	if(WaitForChar(input, 10*100))
	    j = Read(input, (char *)keys, 1);
	else
	    j = 0;
	if(j<1)
	{
	    return -4;
	}
	if(keys[0]==SOH)
	{
	    *size = 128; /* normal block size */
	    break;
	}
	if(keys[0]==STX)
	{
	    *size = 1024; /* 1k block size */
	    break;
	}
	if(keys[0]==EOT)
	    return -3;
	if(keys[0]==CAN)
	    return -2;
    }

    j = 0;
    if(WaitForChar(input, 100))
	j = Read(input, (char *)keys, 2); /* Get block # and its complement */
    if(j==1)
    {
	if(WaitForChar(input, 100))
	    j += Read(input, (char *)keys+1, 1); /* Get block # complement */
    }
    if(j!=2)
    {
	return -4;
    }
    blokki = keys[0];
    if(blokki != 255-keys[1])
    {
	blokki = expblocknum;	/* we can't be sure which one is wrong */
    }

    /* Handle the wrapping in block numbers (only 8 bits are transferred) */
    blokki |= (expblocknum&~255);

    i = 0;
    while(i<*size)
    {
	j = 0;
	if(WaitForChar(input, 100))
	    j = Read(input, (char *)block+i, *size-i);
	if(j>0)
	    i += j;
	else
	    return -4;					/* carrier lost */
    }
    crc = sum = 0;
    for(i=0;i<*size;i++)
    {
	crc = updcrc(block[i], crc);
	sum += block[i];
    }
    sum &= 255;

    if(crcmode)
    {
	j = 0;
	if(WaitForChar(input, 100))
	    j = Read(input, (char *)keys, 1);
	if(j!=1)
	{
	    return -4;
	}
	crc = updcrc(keys[0], crc);
	j = 0;
	if(WaitForChar(input, 100))
	    j = Read(input, (char *)keys+1, 1);
	if(j!=1)
	{
	    return -4;
	}
	crc = updcrc(keys[1], crc);
	if(crc & 0xffff)
	{
	    return -4;
	}
    }
    else
    {
	j = 0;
	if(WaitForChar(input, 100))
	    j = Read(input, (char *)keys, 1);
	
	if(j!=1 || keys[0]!=sum)
	{
	    return -4;
	}
    }
    return blokki;
}


int rx(APTR input, APTR output, char *filename)
{
    unsigned char block[MAXBLOCKSIZE];
    int blocknum, recblock, insize = 0, crcmode;
    int blksize, files = 0, retries;
    FILE *fp = NULL;
    long filesize = -1;
    unsigned long timeval;

    while(1)
    {
	Delay(100);
	retries = 0;
	blocknum = 1;
	crcmode = 0;

	timeval = time(NULL);
	if(0)
	{
	    Write(output, "\025", 1); /* initial NAK */
	}
	else
	{
	    Write(output, "C", 1); /* initial NAK for crc16 */
	    crcmode = 1;
	}
	while(1)
	{
	    recblock = receiveblock(input, output, block, blocknum, &blksize, crcmode);
	    if(recblock==-4)
	    {
		if(WaitForChar(input, 100))
		{
		    /* Clear the input buffer to get in sync better */
		    Read(input, (char *)block, MAXBLOCKSIZE);
		}
		/*fprintf(stderr, "Timeout in receiveblock\n");*/
		if(++retries>MAXRETRIES)
		{
		    recblock = -2;
		    break;
		}
		if(!fp)
		    Write(output, "C", 1); /* Initial NAK for crc16 */
		else
		    Write(output, "\025", 1); /* Some other NAK */
		continue;
	    }
	    if(recblock==-2 || recblock==-3)
		break;

	    if(recblock == 0)
	    {
		if(!block[0])
		{
		    Write(output, "\006", 1); /* ACK the block */
		    return files;
		}

		retries = 0;
		insize = 0;
		blocknum = 1;
		if((fp = fopen((char *)block, "wb")))
		{
		    filesize = atol((char *)block+strlen((char *)block)+1);
		    /*fprintf(stderr, "Receiving \"%s\" (%ld bytes) with ymodem(-batch)\n",
			    block, filesize);*/
		    Write(output, "\006", 1); /* ACK the block */
		    Write(output, "\025", 1); /* NAK to get the first block */
		    /* TODO: for CRC16 send 'C' instead */
		}
		else
		{
		    /*fprintf(stderr, "Could not open '%s' for writing\n",
			    block);*/
		    Write(output, "\030\030", 2); /* CAN CAN */
		    /*printf("\nTransfer aborted.\n\r");*/
		    return files;
		}
		timeval = time(NULL);
	    }
	    else if(recblock >= 0)
	    {
		if(recblock == blocknum)
		{
		    if(!fp)
		    {
			/* This is then Xmodem..
			   Use the filename given to us. */
			fp = fopen(filename, "wb");
			if(!fp)
			{
			    /*fprintf(stderr,
				    "Could not open \"%s\" for writing\n",
				    filename);*/
			    Write(output, "\030\030", 2); /* CAN CAN */
			    /*printf("\nTransfer aborted.\n\r");*/
			    return files;
			}
			/*fprintf(stderr,
				"Receiving \"%s\" with xmodem\n",
				block);*/
		    }
		    /* if right block, save it */
		    if((filesize<0) || (filesize-insize >= blksize))
		    {
			if(fp)
			    fwrite(block, blksize, 1, fp);
			insize += blksize;
		    }
		    else
		    {	
			if(fp)
			    fwrite(block, filesize-insize, 1, fp);
			insize = filesize;
		    }
		    blocknum++;
		    Write(output, "\006", 1); /* ACK the block */
		    if(retries)
			--retries;
		}
		else if(recblock < blocknum)
		{
		    Write(output, "\006", 1); /* ACK if we are ahead */
		}
		else
		{
		    if(fp)
		    {
			fclose(fp);
		    }
		    Write(output, "\030\030", 2); /* CAN CAN */
		    /*printf("\nTransfer aborted.\n\r");*/
		    return files;
		}
	    }
	    else
	    {
		/* if block is bad, but we are ahead, ACK it */
		if(recblock<blocknum)
		{
		    Write(output, "\006", 1);
		}
		else
		{
		    /* if we need this block or can't make it clear, NAK it */

		    if(WaitForChar(input, 100))
		    {
			/* Clear the input buffer to get in sync better */
			Read(input, (char *)block, MAXBLOCKSIZE);
		    }
		    Write(output, "\025", 1); /* NAK */
		    retries++;
		}
	    }
	    /*fprintf(stderr, "Block: %d Bytes: %d\r", recblock, insize);*/
	}
	if(recblock==-2)
	{
	    if(fp)
		fclose(fp);
	    Write(output, "\030\030", 2); /* CAN CAN */
	    /*printf("\nTransfer aborted.\n\r");*/
	    return files;
	}
	Write(output, "\006", 1);	/* acknowledge the EOT */
	fclose(fp);
	fp = NULL;
	files++;
	retries = 0;

	/*
	timeval = time(NULL)-timeval;
	if(timeval)
	    printf("\nTransfer successful. (%d cps)\n", insize/timeval);
	else
	    printf("\nTransfer successful.\n");
	return 0;
	*/
    }
    return files;
}


int main(int argc,char *argv[])
{
    char *filename = "xmodem-tmp";
    LONG error;

    if(argc==2)
    {
	filename = argv[1];
    }
    SetMode(Input(), MODE_RAW|MODE_NOCRLF);
    SetMode(Output(), MODE_RAW|MODE_NOCRLF);
    error = rx(Input(), Output(), filename);
    SetMode(Input(), MODE_CONSOLE);
    SetMode(Output(), MODE_CONSOLE);

    if(error)
    {
	printf("\b\n%d file%s received\n", error, (error>1)?"s":"");
	return 0;
    }
    return 20;
}
