#include <exec/types.h>
#include <exec/memory.h>
#include <exec/exec_protos.h>
#include <dos/dos_protos.h>
#include <stdio.h>
#include <stdlib.h>

#include "crctab.c"

/* send xmodem/ymodem batch */

#define NUL 0x00
#define SOH 0x01
#define STX 0x02
#define EOT 0x04
#define ACK 0x06
#define NAK 0x15
#define CAN 0x18

#define MAXRETRIES 10

void __chkabort(void) {}
int getkey(unsigned long);
int getblock(APTR handle, char *block, int *size);
void sendblock(char *block, unsigned long blocknum, int size, int crcmode);

#define MAXBLOCKSIZE 1024

int main(int argc,char *argv[])
{
    int c, n;
    int retries = 0, firstblock = 1, blocksize = 128, crcmode = 0;
    static char block[MAXBLOCKSIZE];
    char *filename, *tmp;
    int blocknum, lastblock = 0, files = 0;
    APTR handle;
    ULONG size = 0;

    if(argc<2)
    {
	printf("Usage: 'sx filename'\n\r");
	return RETURN_WARN;
    }

    printf("***Sending using xmodem-chksum. Use ctrl-X's to cancel\n\r");
    SetMode(Output(), MODE_RAW | MODE_NOCRLF);
    SetMode(Input(), MODE_RAW | MODE_NOCRLF);

    for(n=1;n<=argc;n++) /* One extra round */
    {
	handle = NULL;
	size = 0;
	if(n<argc)
	{
	    if((handle = Lock(argv[n], ACCESS_READ)))
	    {
		struct FileInfoBlock fib;

		if(Examine(handle, &fib))
		{
		    size = fib.fib_Size;
		}
		UnLock(handle);
	    }
	    if(!(handle = Open(argv[n],MODE_OLDFILE)))
		continue;
	}
	retries = 0;
	firstblock = 1;
	lastblock = 0;
	crcmode = 0;
	blocksize = 128;
#if 0
	blocknum = 1;
	if(getblock(handle, block, &blocksize))
	{
	    lastblock = 1;
	}
#else
	for(c=0;c<blocksize;c++)
	{
	    block[c] = 0;
	}
	if(handle)
	{
	    tmp = argv[n];

	    /* Strip the path from the filename */
	    filename = tmp;
	    while(*tmp)
	    {
		if(*tmp=='/' || *tmp==':')
		    filename = tmp+1;
		tmp++;
	    }
	    sprintf(block, "%s%c%d", filename, 0, size);
	}
	blocknum = 0;
#endif

	while(1)
	{
	    while(retries<MAXRETRIES)
	    {
		c = getkey(1000);
		if(firstblock)
		{
		    /* block 0 ACK'ed, wait for start of transfer */
		    if(c==ACK)
		    {
			if(!handle)
			{
			    break;
			}
			
			firstblock = 0;
			blocknum = 1;
			blocksize = 1024;
			if(getblock(handle, block, &blocksize))
			{
			    lastblock = 1;
			}
		    }
		    if(c=='C')
		    {
			crcmode = 1;
			c = NAK;
		    }
		    if(c==NAK)
		    {
			sendblock(block, blocknum, blocksize, crcmode);
		    }
		    ++retries;
		}
		else
		{
		    if(c==ACK)
			break;
		    if(c==CAN)
		    {
			retries += 2;
		    }
		    if(c==NAK)
		    {
			/* TODO: if retries >0, blocknum==1 */
			/* fall back to 128-byte blocks */
			sendblock(block, blocknum, blocksize, crcmode);
			firstblock = 0;
		    }
		    ++retries;
		}
	    }
	    if(!handle)
		break;

	    if(retries>=MAXRETRIES)
	    {
		retries = MAXRETRIES;
		break;
	    }
	    if(lastblock)
	    {
		Write(Output(), "\004", 1); /* EOT */
		firstblock = 0;
		break;
	    }
	    blocksize = 1024;
	    if(getblock(handle, block, &blocksize))
	    {
		lastblock = 1;
		Write(Output(), "\004", 1); /* EOT */
		firstblock = 0;
		break;
	    }
	    blocknum++;
	    sendblock(block, blocknum, blocksize, crcmode);
	    firstblock = 0;
	}
	if(handle)
	{
	    if(retries>=MAXRETRIES)
	    {
		Close(handle);
		Write(Output(), "\030\030", 2); /* CAN CAN */
		SetMode(Input(), MODE_CONSOLE);
		SetMode(Output(), MODE_CONSOLE);
		return 20;
	    }
	    retries=0;
	    while(1)
	    {
		c = getkey(1000);
		if(c==ACK)
		    break;
		if(c==CAN)
		{
		    retries += 2;
		}
		if(c==NAK)
		{
		    Write(Output(), "\004", 1); /* EOT */
		}
		++retries;
	    }
	    if(retries<MAXRETRIES)
	    {
		files++;
	    }
	    /*
	       if(retries>=MAXRETRIES)
	       fprintf(stderr, "\nFinal ACK missing.\n");
	       else
	       {
	       fprintf(stderr, "\nTransfer successful (%d blocks).\n", blocknum);
	       }
	       */
	    Close(handle);
	}
    }
    SetMode(Output(), MODE_CONSOLE);
    SetMode(Input(), MODE_CONSOLE);
    if(files)
	return RETURN_OK;
    else
	return RETURN_ERROR;
}

int getkey(unsigned long time)
{
    char q;

    if(WaitForChar(Input(), time))
    {
	if(Read(Input(), &q, 1))
	    return q;
    }
    return -1;
}


/* If blocksize==1024, it may be set to 128 for the last block in file. */
/* If blocksize==128, it may NOT be changed */
int getblock(APTR handle, char *block, int *blocksize)
{
    int i;

    i = Read(handle, block, *blocksize);
    if(i<1)
    {
	return -1; /* No more data */
    }
    if(i<*blocksize)
    {
	if(i<=128)
	    *blocksize = 128;
	while(i<*blocksize)
	    block[i++] = '\0';
    }
    return 0;
}


void sendblock(char *block, unsigned long blocknum, int size, int crcmode)
{
    unsigned char q[3];
    unsigned long crc = 0, sum = 0;
    int i;

    for(i=0;i<size;i++)
    {
	crc = updcrc(block[i], crc);
	sum += block[i];
    }

    if(size==128)
	q[0] = SOH;
    else
	q[0] = STX;
    q[1] = (blocknum&255);
    q[2] = 255-(blocknum&255);
    Write(Output(), (char *)q, 3);
    Write(Output(), block, size);

    if(crcmode)
    {
	crc = updcrc(0, updcrc(0, crc));
	q[0] = (crc>>8);
	q[1] = (crc&255);
	Write(Output(), (char *)q, 2);
    }
    else
    {
	q[0] = sum;
	Write(Output(), (char *)q, 1);
    }
}

