#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>

/*
   pmtopgm.c v1.0 1994-10-12 by Henrik Herranen
   e-mail: leopold@cs.tut.fi
*/
#if 0
int printf(char *, ...);
int fprintf(FILE *, char *, ...);
int fread(char *, int, int, FILE *);
int fwrite(char *, int, int, FILE *);
int fseek(FILE *, int, int);
void fclose(FILE *);
#endif

static int HandleImage(FILE *InFile, FILE *OutFile);

int main(int argc, char **argv) {
  static FILE *InFile, *OutFile;

  if (argc>3) {
    fprintf(stderr, "%s: too many parameters\nusage: %s [InFile [OutFile]]\n",
	    argv[0], argv[0]);
    return 10;
  }

  if (argc>1) {
    if (!(InFile = fopen(argv[1], "rb"))) {
      fprintf(stderr, "%s: couldn't open read file %s\n", argv[0], argv[1]);
      return 10;
    }
    if (argc>2) {
      if (!(OutFile = fopen(argv[2], "wb"))) {
	fprintf(stderr, "%s: couldn't open write file %s\n", argv[0], argv[2]);
	return 10;
      }
      return HandleImage(InFile, OutFile);
    } else {
      return HandleImage(InFile, stdout);
    }
  } else {
    return HandleImage(stdin, stdout);
  }
}


struct pmHeader {
  unsigned char magic[4];
  short version;
  short xsize, ysize;
  short bitsperpixel;
  long bodysize;
};

static int HandleImage(FILE *InFile, FILE *OutFile) {
  static struct pmHeader Head;
  static int width, height, bits, version, bodysize, bytesperrow;
  static unsigned char *body, *line;
  static int x, y;
  static int color[4] = {248, 255, 192, 0};

  if (fread((char *)&Head, sizeof(struct pmHeader), 1, InFile) != 1 ||
      Head.magic[0] != 0x61 ||
      Head.magic[1] != 0x0a ||
      Head.magic[2] != 0x59 ||
      Head.magic[3] != 0x26) {
    fprintf(stderr, "improper header\n");
    return 10;
  }

  width = ntohs(Head.xsize);
  height = ntohs(Head.ysize);
  version = ntohs(Head.version);
  bits = ntohs(Head.bitsperpixel);
  bodysize = ntohl(Head.bodysize);
  bytesperrow = ((width + 15) & ~15) / 4;

  if (version != 0) {
    fprintf(stderr, "can only handle version 0 files (this was %d)\n", version);
    return 10;
  }

  if (bits != 2) {
    fprintf(stderr, "can only handle 2 bitplane files (this was %d)\n", bits);
    return 10;
  }

  if (bytesperrow*height != bodysize) {
    fprintf(stderr, "expected %d byte body, header implies %d bytes!\n",
	    bytesperrow*height, bodysize);
    return 10;
  }

  if (!(body = malloc(bodysize))) {
    fprintf(stderr, "couldn't allocate %d bytes of memory!", bodysize);
    return 10;
  }

  if (!(line = malloc(width))) {
    fprintf(stderr, "couldn't allocate %d bytes of memory!", width);
    return 10;
  }

  if (fread((char *)body, bodysize, 1, InFile) != 1) {
    fprintf(stderr, "unexpected end of file!\n");
    return 10;
  }

  fprintf(OutFile, "P5\n%d %d\n255\n", width, height);
  for(y=0; y<height; y++) {
    for (x=0; x<width; x++) {
      line[x] = color[(body[y*bytesperrow + (x>>2)] >> ((3-(x&3))<<1)) & 3];
    }
    fwrite((char *)line, width, 1, OutFile);
  }

  return 0;
}

